"""Fermionic symmetric arrays with flat backend."""

import functools

import autoray as ar

from ..abelian_common import AbelianCommon
from ..common import SymmrayCommon
from ..fermionic_common import FermionicCommon
from ..fermionic_local_operators import FermionicOperator
from ..sparse.sparse_fermionic_array import FermionicArray
from ..symmetries import get_symmetry
from ..utils import DEBUG
from .flat_array_common import FlatArrayCommon, truncate_svd_result_flat
from .flat_data_common import FlatCommon
from .flat_vector import FlatVector


@functools.lru_cache(maxsize=2**14)
def perm_to_swaps(perm):
    """Convert a permutation to a list of swaps.

    Parameters
    ----------
    perm : tuple[int, ...]
        The permutation to convert. Should be a permutation of the elements in
        range(len(perm)).

    Returns
    -------
    tuple[tuple[int, int], ...]
        A tuple of swaps that will perform the permutation.
    """
    N = len(perm)
    current = list(range(N))
    desired = list(map(int, perm))
    idxs = {i: i for i in current}
    swaps = []

    for pos in range(N):
        x = desired[pos]
        ix = idxs[x]
        # ix = current.index(x)
        in_position = ix == pos
        while not in_position:
            # swap with item to left
            iy = ix - 1
            y = current[iy]
            # record the swap
            swaps.append((iy, ix))
            current[iy], current[ix] = x, y
            idxs[y], idxs[x] = ix, iy
            # check if we have finished this element
            ix = iy
            in_position = ix == pos

    return tuple(swaps)


def oddpos_parse(oddpos, parity):
    if oddpos is None:
        # no dummy indices
        return (), ()

    if isinstance(oddpos, (list, tuple)) and isinstance(
        oddpos[0], FermionicOperator
    ):
        # an explicit sequence of subsumed odd ranks
        oddpos = tuple(oddpos)

        if not isinstance(parity, (list, tuple)):
            parities = (parity,) * len(oddpos)
        else:
            parities = tuple(parity)

        return oddpos, parities

    if not isinstance(oddpos, FermionicOperator):
        oddpos = FermionicOperator(oddpos)

    return (oddpos,), (parity,)


class FermionicArrayFlat(
    FermionicCommon,
    FlatArrayCommon,
    FlatCommon,
    AbelianCommon,
    SymmrayCommon,
):
    """Fermionic abelian symmetric array with flat backend.

    Parameters
    ----------
    sectors : array_like
        The stack of sector keys, with shape (num_blocks, ndim). Each row
        represents a sector of a corresponding block, and each column
        represents a charge in a given axis.
    blocks : array_like
        The stack of array blocks, with shape (num_blocks, *shape_block), i.e.
        `ndim + 1` dimensions, where the first dimension is the block index,
        which should match the first dimension of `sectors`, and the rest are
        the dimensions of individual blocks.
    indices : sequence[FlatIndex]
        Indices describing the dualness and any subindex information for each
        dimension of the array. If bools are supplied, they will be converted
        to a FlatIndex with the corresponding dualness, and no subindex
        information.
    phases : array_like, optional
        An array of +/- 1 phases, with shape (num_blocks,), giving the phase
        of each block. If not supplied, all phases are assumed to be +1.
    label : hashable, optional
        An optional label for the array, potentially needed for ordering dummy
        odd fermionic modes.
    symmetry : str or Symmetry, optional
        The symmetry of the array, if not using a specific symmetry class.
    """

    __slots__ = (
        "_blocks",
        "_indices",
        "_odd_parities",
        "_oddpos",
        "_phases",
        "_sectors",
        "_symmetry",
        "backend",
    )
    fermionic = True
    static_symmetry = None

    def __init__(
        self,
        sectors,
        blocks,
        indices,
        phases=None,
        label=None,
        symmetry=None,
        oddpos=None,
        odd_parities=None,
    ):
        self._init_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
            symmetry=symmetry,
            label=label,
        )

        if phases is None:
            self._phases = None
        elif hasattr(phases, "shape"):
            self._phases = phases
        else:
            self._phases = ar.do("array", phases, like=self._blocks)

        if oddpos is None and self.label is not None:
            # default to the array label
            oddpos = self.label

        if oddpos:
            oddpos, odd_parities = oddpos_parse(oddpos, self.parity)
        else:
            # None or empty
            oddpos = ()
            odd_parities = ()

        self._oddpos = oddpos
        self._odd_parities = odd_parities

        if DEBUG:
            self.check()

    @property
    def phases(self):
        """The phases for each block."""
        if self._phases is None:
            self._phases = ar.do("ones", (self.num_blocks,), like=self._blocks)
        return self._phases

    @property
    def oddpos(self):
        """Any labels of dummy fermionic modes for odd parity tensors."""
        return self._oddpos

    @property
    def odd_parities(self):
        """The parities of each oddpos label carried by this array."""
        return self._odd_parities

    @property
    def parity(self):
        """The total parity of the array, 0 for even, 1 for odd."""
        return ar.do("sum", self.sectors[0] % 2, like=self.backend) % 2

    def check(self):
        """Check the internal consistency of the array."""
        self._check_abelian()
        if self._phases is not None:
            assert ar.do("shape", self._phases) == (self.num_blocks,)
            assert ar.do("all", ar.do("isin", self._phases, [-1, 1]))
        assert len(self._oddpos) == len(self._odd_parities)

    def new_with(self, sectors, blocks, indices) -> "FermionicArrayFlat":
        """Create a new flat fermionic array of the same class as this one.
        Unlike `copy`, this does not copy over any existing data and drops
        for example `label`, `phases`, and `oddpos`.
        """
        new = self._new_with_abelian(sectors, blocks, indices)
        new._phases = None
        new._oddpos = ()
        new._odd_parities = ()
        return new

    def copy(self, deep=False) -> "FermionicArrayFlat":
        """Create a copy of the array."""
        new = self._copy_abelian(deep=deep)

        if self._phases is not None:
            if deep:
                new._phases = ar.do("copy", self._phases, like=self.backend)
            else:
                new._phases = self._phases
        else:
            new._phases = None

        new._oddpos = self._oddpos
        new._odd_parities = self._odd_parities
        return new

    def copy_with(
        self,
        sectors=None,
        blocks=None,
        indices=None,
        phases=None,
    ) -> "FermionicArrayFlat":
        """A copy of this fermionic flat array with some attributes replaced.
        Note that checks are not performed on the new properties, this is
        intended for internal use.
        """
        new = self._copy_with_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
        )
        new._phases = self._phases if phases is None else phases
        new._oddpos = self._oddpos
        new._odd_parities = self._odd_parities
        if DEBUG:
            new.check()
        return new

    def modify(
        self,
        sectors=None,
        blocks=None,
        indices=None,
        phases=None,
        oddpos=None,
        odd_parities=None,
    ) -> "FermionicArrayFlat":
        """Modify this fermionic flat array in place with some attributes
        replaced. Note that checks are not performed on the new properties,
        this is intended for internal use.
        """
        self._modify_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
        )
        if phases is not None:
            if isinstance(phases, str):
                if phases == "reset":
                    self._phases = None
                else:
                    raise ValueError(f"Unknown phases value '{phases}'")
            else:
                self._phases = phases

        if oddpos is not None:
            self._oddpos = oddpos
        if odd_parities is not None:
            self._odd_parities = odd_parities
        if DEBUG:
            self.check()
        return self

    def set_params(self, params):
        """Set the underlying array blocks."""
        self._set_params_abelian(params)

        if self._phases is not None:
            try:
                self._phases = ar.do("asarray", self._phases, like=params)
            except ImportError:
                # params is possibly a placeholder of some kind
                pass

        if DEBUG:
            self.check()

    def _map_blocks(self, fn_sector=None, fn_block=None):
        self._map_blocks_abelian(fn_sector=fn_sector, fn_block=fn_block)
        if fn_sector is not None:
            if self._phases is not None:
                # NOTE: leave missing phases, assumed to stay all ones
                self._phases = fn_sector(self.phases)

    @classmethod
    def from_blocks(
        cls,
        blocks,
        indices,
        phases=None,
        oddpos=None,
        symmetry=None,
        label=None,
    ) -> "FermionicArrayFlat":
        """Create a fermionic flat array from an explicit dictionary of blocks,
        and sequence of indices or duals.

        Parameters
        ----------
        blocks : dict[tuple[int, ...], array_like]
            A dictionary mapping sector keys (tuples of charges) to blocks
            (arrays).
        indices : sequence[FlatIndex] | sequence[bool]
            A sequence of indices describing the dualness and any subindex
            information for each dimension of the array. If bools are supplied,
            they will be converted to a FlatIndex with the corresponding
            dualness, and no subindex information.
        phases : dict[tuple[int, ...], int], optional
            A dictionary mapping sector keys to +/- 1 phases. If not supplied,
            all phases are assumed to be +1.
        symmetry : str or Symmetry, optional
            The symmetry of the array, if not using a specific symmetry class.
        label : hashable, optional
            An optional label for the array, potentially needed for ordering
            dummy odd fermionic modes.

        """
        sectors = list(map(list, blocks.keys()))

        if phases is not None:
            phases = ar.do(
                "array",
                [phases.get(sector, 1) for sector in blocks.keys()],
                like=blocks,
            )

        if blocks:
            blocks = ar.do("stack", tuple(blocks.values()))
        else:
            blocks = []

        return cls(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
            phases=phases,
            oddpos=oddpos,
            symmetry=symmetry,
            label=label,
        )

    @classmethod
    def from_blocksparse(
        cls,
        x: "FermionicArray",
        symmetry=None,
    ):
        """Create a fermionic flat array from a fermionic blocksparse array.

        Parameters
        ----------
        x : FermionicArray
            The fermionic blocksparse array to convert.
        symmetry : str or Symmetry, optional
            The symmetry to use. If not supplied, the symmetry of `x` is used.
        """
        return cls.from_blocks(
            blocks=x.blocks,
            indices=[ix.to_flat() for ix in x.indices],
            phases=x.phases,
            oddpos=x.oddpos,
            symmetry=symmetry or x.symmetry,
            label=x.label,
        )

    def to_blocksparse(self) -> "FermionicArray":
        """Convert this fermionic flat array to a fermionic blocksparse array.

        Returns
        -------
        FermionicArray
            The equivalent fermionic blocksparse array.
        """
        if self._phases is None:
            phases = {}
        else:
            phases = {
                tuple(map(int, sector)): int(phase)
                for sector, phase in zip(self._sectors, self._phases)
                if phase != 1
            }

        oddpos = tuple(
            f for f, p in zip(self._oddpos, self._odd_parities) if p == 1
        )

        return self._to_blocksparse_abelian(phases=phases, oddpos=oddpos)

    def sort_stack(
        self,
        axes=None,
        all_axes=None,
        inplace=False,
    ) -> "FermionicArrayFlat":
        """Lexicgraphic sort the stack of sectors and blocks according to the
        values of charges in the specified axes, optionally filling in the rest
        of the axes with the remaining axes in the order they appear.

        Parameters
        ----------
        axes : int | tuple[int, ...], optional
            The axes to sort by. If a single integer is given, it will be
            interpreted as the axis to sort by. If a tuple of integers is
            given, it will be interpreted as the axes to sort by in order.
            Default is None, if all_axes is also None or True, this will sort
            all axes in their current order.
        all_axes : bool, optional
            Whether to include all non-specified axes as tie-breakers, after
            the specified axes. If ``None``, the default, this will be True
            if `axes` is not supplied explicitly, and False otherwise.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.
            Default is False, which returns a new array.
        """
        if self.ndim <= 1:
            # nothing to do
            return self if inplace else self.copy()

        kord = self.get_sorting_indices(axes=axes, all_axes=all_axes)

        new_sectors = self._sectors[kord]
        new_blocks = self._blocks[kord]
        if self._phases is None:
            new_phases = None
        else:
            new_phases = self._phases[kord]

        return self._modify_or_copy(
            sectors=new_sectors,
            blocks=new_blocks,
            phases=new_phases,
            inplace=inplace,
        )

    def phase_sync(self, inplace=False) -> "FermionicArrayFlat":
        """Multiply all lazy phases into the block arrays.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArrayFlat
            The resolved array, which now has all trivial (+1) phases.
        """
        new = self if inplace else self.copy()

        if new._phases is not None:
            # do broadcasted multiply to resolve phases
            phases_b = new._phases[(slice(None),) + (None,) * (new.ndim)]
            new.modify(
                blocks=new._blocks * phases_b,
                phases="reset",
            )

        return new

    def phase_flip(self, *axs, inplace=False) -> "FermionicArrayFlat":
        """Flip the phase of all sectors with odd parity at the given axis.

        Parameters
        ----------
        ax : int
            The axis along which to flip the phase.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArrayFlat
            The phase-flipped array.
        """
        new = self if inplace else self.copy()
        if not axs:
            # nothing to do
            return new
        flip_phases = (ar.do("sum", new._sectors[:, axs], axis=1) % 2) * -2 + 1
        new.modify(phases=new.phases * flip_phases)
        return new

    def phase_transpose(
        self,
        axes=None,
        inplace=False,
    ) -> "FermionicArrayFlat":
        """Phase this fermionic array as if it were transposed virtually, i.e.
        the actual arrays are not transposed. Useful when one wants the actual
        data layout to differ from the required fermionic mode layout.

        Parameters
        ----------
        axes : tuple of int, optional
            The permutation of axes, by default None.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArrayFlat
        """
        new = self if inplace else self.copy()

        N = new.ndim
        if axes is None:
            # full reversal, shortcut to count the swaps
            nswap = (
                ar.do("sum", new.sectors % 2, axis=1, like=self.backend) // 2
            )
        elif all(ax == i for i, ax in enumerate(axes)):
            # identity, nothing to do
            return new
        else:
            # convert permutation to sequence of pairwise neighboring swaps
            swaps = perm_to_swaps(tuple(axes))

            # count how many swaps of odd charges there are
            parities = [new.sectors[:, i] % 2 for i in range(N)]
            nswap = None
            for il, ir in swaps:
                if nswap is None:
                    nswap = parities[il] * parities[ir]
                else:
                    nswap = nswap + (parities[il] * parities[ir])
                parities[il], parities[ir] = parities[ir], parities[il]

        if nswap is None:
            raise ValueError("No phase changes required.")

        # absorb into current phases
        phase_change = (nswap % 2) * -2 + 1
        new.modify(
            phases=(
                phase_change
                if new._phases is None
                else new._phases * phase_change
            )
        )

        return new

    def phase_sector(self, sector, inplace=False):
        raise NotImplementedError(
            "Flat fermionic arrays do not support individual sector phasing."
        )

    def phase_global(self, inplace=False):
        """Flip the global phase of the array.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArray
        """
        new = self if inplace else self.copy()
        new.modify(phases=-new.phases)
        return new

    def _resolve_oddpos_conj(self, phase_permutation=True):
        """Assuming we have effectively taken the conjugate of a fermionic
        array with dummy oddpos modes, get their new order and compute any
        phase changes coming from moving back to the beginning of the index
        order.
        """
        if not self.oddpos:
            return

        # 1. we get a reversal and conjugation of the oddpos modes
        #       dummy modes          real indices
        # | o0 o1 ... on-2 on-1 | P0 P1 ... Pn-2 Pn-1 |
        #                     <-->
        # | Pn-1 Pn-2 ... P1 P0 | on-1 on-2 ... o1 o0 |
        new_oddpos = tuple(r.dag for r in reversed(self.oddpos))
        new_odd_parities = tuple(reversed(self.odd_parities))
        self.modify(oddpos=new_oddpos, odd_parities=new_odd_parities)

        if phase_permutation:
            # 2. moving oddpos charges back to left
            # after flipping might generate global sign
            # | Pn-1 Pn-2 ... P1 P0 | on-1 on-2 ... o1 o0 |
            #                     <--
            # | on-1 on-2 ... o1 o0 | Pn-1 Pn-2 ... P1 P0 |
            sign = (self.parity * sum(self.odd_parities) % 2) * -2 + 1
            self.modify(phases=self.phases * sign)

    def _resolve_oddpos_combine(self, a, b):
        """Calculate the new combined dummy odd modes and any associated global
        phases combing from contracting two fermionic arrays `a` and `b`. This
        modifies this array in place.
        """
        l_oddpos = a.oddpos
        r_oddpos = b.oddpos
        if not l_oddpos and not r_oddpos:
            # no odd modes to resolve
            self._oddpos = self._odd_parities = ()
            return

        l_odd_parities = a.odd_parities
        r_odd_parities = b.odd_parities

        oddpos = [*l_oddpos, *r_oddpos]
        odd_parities = [*l_odd_parities, *r_odd_parities]

        # 1. initially we have:
        # left dummy modes | left real modes | right dummy modes | right real modes
        # so we must calc phase from moving right dummy modes past left real modes
        phase = (a.parity * sum(r_odd_parities) % 2) * -2 + 1

        # then we want to sort the joint set of left and right dummy modes,
        perm = tuple(sorted(range(len(oddpos)), key=lambda i: oddpos[i]))
        swaps = perm_to_swaps(perm)
        for i, j in swaps:
            a, b = oddpos[i], oddpos[j]
            pa, pb = odd_parities[i], odd_parities[j]
            # compute sign from swap
            phase = phase * ((pa * pb) * -2 + 1)
            # perform swap
            oddpos[i], oddpos[j] = b, a
            odd_parities[i], odd_parities[j] = pb, pa

        # do the global phase, and set the new sorted oddpos labels and parities
        self.modify(
            oddpos=tuple(oddpos),
            odd_parities=tuple(odd_parities),
            phases=self.phases * phase,
        )

    def _resolve_oddpos_squeeze(self, axes_squeeze):
        """Assuming we are about to squeeze away `axes_squeeze`, compute the
        phases associated with moving them to the beginning of the array, and
        then turn them into dummy oddpos modes.
        """
        axes_leave = []
        squeezed_oddpos = []
        squeezed_odd_parities = []
        for ax, ix in enumerate(self.indices):
            if ax in axes_squeeze:
                # all sectors should have same charge for squeezed axes
                odd_parity = self._sectors[0, ax] % 2

                label = self.label
                if label is None:
                    raise ValueError(
                        "Cannot squeeze flat fermionic index with possible odd"
                        " parity if array has no ordering `.label` attribute."
                    )
                op = FermionicOperator(("squeeze", label, ax), ix.dual)
                squeezed_oddpos.append(op)
                squeezed_odd_parities.append(odd_parity)
            else:
                axes_leave.append(ax)

        self.phase_transpose((*axes_squeeze, *axes_leave), inplace=True)
        self.modify(
            oddpos=(*self.oddpos, *squeezed_oddpos),
            odd_parities=(*self.odd_parities, *squeezed_odd_parities),
        )

    def transpose(
        self,
        axes=None,
        phase=True,
        inplace=False,
    ) -> "FermionicArrayFlat":
        """Transpose this flat fermionic array, by default accounting for the
        phases accumulated from swapping odd charges.

        Parameters
        ----------
        axes : tuple of int, optional
            The permutation of axes, by default None.
        phase : bool, optional
            Whether to flip the phase of sectors whose odd charges undergo a
            odd permutation. By default True.
        inplace : bool, optional
            Whether to perform the operation in place.

        Returns
        -------
        FermionicArrayFlat
            The transposed array.
        """
        new = self if inplace else self.copy()

        if phase:
            # perform the phase accumulation separately first
            new.phase_transpose(axes, inplace=True)

        # transpose the actual arrays
        return new._transpose_abelian(axes, inplace=True)

    def conj(self, phase_permutation=True, phase_dual=False, inplace=False):
        """Conjugate this flat fermionic array. By default this include phases
        from both the virtual flipping of all axes, but *not* the conjugation
        of dual indices, such that::

            (
                tensordot_fermionic(x.conj(), x, ndim) ==
                tensordot_fermionic(x, x.conj(), ndim)
            )

        If all indices have matching dualness (i.e. all bra or all ket), *or*
        you set `phase_dual=True` then the above contractions will also be
        equal to ``x.norm() ** 2``.

        Parameters
        ----------
        phase_permutation : bool, optional
            Whether to flip the phase of sectors whose odd charges undergo a
            odd permutation due to *virtually* flipping the order of axes, by
            default True.
        phase_dual : bool, optional
            Whether to flip the phase of dual indices, by default False. If a
            FermionicArrayFlat has a mix of dual and non-dual indices, and you
            are explicitly forming the norm, you may want to set this to True.
            But if it is part of a large tensor network you only need to flip
            the phase of true 'outer' dual indices.

        Returns
        -------
        FermionicArrayFlat
        """
        new = self if inplace else self.copy()

        if phase_permutation:
            # perform the phase accumulation separately first
            new.phase_transpose(inplace=True)

        # conjugate the actual arrays
        new._conj_abelian(inplace=True)

        if phase_dual:
            axs_conj = tuple(
                ax for ax, ix in enumerate(new.indices) if not ix.dual
            )
            new.phase_flip(*axs_conj, inplace=True)

        if new.oddpos:
            # handle potential dummy odd modes
            new._resolve_oddpos_conj(phase_permutation)

        return new

    def dagger(self, phase_dual=False, inplace=False):
        """Fermionic conjugate transpose."""
        new = self._conj_abelian(inplace=inplace)
        new._transpose_abelian(inplace=True)

        if phase_dual:
            axs_conj = tuple(
                ax for ax, ix in enumerate(new.indices) if not ix.dual
            )
            new.phase_flip(*axs_conj, inplace=True)

        # handle potential dummy odd modes
        #     dagger defined by phaseless reversal of all axes
        new._resolve_oddpos_conj(phase_permutation=True)

        return new

    # ------------------------------- linalg -------------------------------- #

    def eigh(
        self,
        phase_eigenvalues=True,
    ) -> tuple[FlatVector, "FermionicArrayFlat"]:
        """Hermitian eigen-decomposition of this flat fermionic array.

        Parameters
        ----------
        phase_eigenvalues : bool, optional
            If True, any local phase will be absorbed into the eigenvalues,
            such that `U @ diag(w) @ U.H == a` always holds. By default
            True. If `False` then one of `U` or `U.H` should be phased
            individually to account for local phasesin the above expression.

        Returns
        -------
        eigenvalues : FlatVector
            The eigenvalues.
        eigenvectors : FermionicArrayFlat
            The abelian array of right eigenvectors.
        """
        x = self.phase_sync()

        w, U = x._eigh_abelian()

        if phase_eigenvalues and not x.indices[1].dual:
            # inner index is like |x><x| so introduce a phase flip,
            # we don't explicitly have Wdag so put phase in eigenvalues
            # XXX: is this the most compatible thing to do?
            # it means U @ diag(w) @ U.H == x always
            parities = w._sectors % 2
            w.modify(blocks=w._blocks * (parities * -2 + 1)[:, None])

        return w, U

    def eigh_truncated(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        positive=0,
        drop_oddpos=True,
        **kwargs,
    ) -> tuple["FermionicArrayFlat", FlatVector, "FermionicArrayFlat"]:
        """Truncated hermitian eigen-decomposition of this assumed hermitian
        flat fermionic array.

        Parameters
        ----------
        cutoff : float, optional
            Absolute eigenvalue cutoff threshold.
        cutoff_mode : int or str, optional
            How to perform the truncation:

            - 1 or 'abs': trim values below ``cutoff``
            - 2 or 'rel': trim values below ``s[0] * cutoff``
            - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

        max_bond : int
            An explicit maximum bond dimension, use -1 for none.
        absorb : {-1, 0, 1, None}
            How to absorb the eigenvalues.

            - -1 or 'left': absorb into the left factor (U).
            - 0 or 'both': absorb the square root into both factors.
            - 1 or 'right': absorb into the right factor (VH).
            - None: do not absorb, return eigenvalues as a BlockVector.

        renorm : {0, 1}
            Whether to renormalize the eigenvalues (depends on `cutoff_mode`).
        positive: bool, optional
            If True, assume all eigenvalues are positive for a faster sort.
            By default False.
        drop_oddpos : bool, optional
            Whether to drop any dummy oddpos modes after the decomposition.
            By default True.

        Returns
        -------
        u : FermionicArrayFlat
            The fermionic array of left eigenvectors.
        w : FlatVector or None
            The vector of eigenvalues, or None if absorbed.
        uh : FermionicArrayFlat
            The fermionic array of right eigenvectors.
        """
        if kwargs:
            import warnings

            warnings.warn(
                f"Got unexpected kwargs {kwargs} in eigh_truncated "
                f"for {self.__class__}. Ignoring them.",
                UserWarning,
            )

        # since we are handling UH, we can add phase there
        w, U = self.eigh(phase_eigenvalues=False)

        # make sure to sort by descending absolute value
        if not positive:
            idx = ar.do(
                "argsort", -ar.do("abs", w._blocks, like=self.backend), axis=1
            )
            w.modify(
                blocks=ar.do("take_along_axis", w._blocks, idx, axis=1),
            )
            U.modify(
                blocks=ar.do(
                    "take_along_axis", U._blocks, idx[:, None, :], axis=2
                )
            )
        else:
            # assume all positive, just need to flip
            w.modify(blocks=ar.do("flip", w._blocks, 1))
            U.modify(blocks=ar.do("flip", U._blocks, 2))

        if DEBUG:
            w.check()
            U.check()

        if drop_oddpos:
            U._oddpos = ()
            U._odd_parities = ()
            U._label = None

        VH = U._dagger_abelian()

        if VH.indices[0].dual:
            # inner index is like |x><x| so introduce a phase flip
            VH.phase_flip(0, inplace=True)

        return truncate_svd_result_flat(
            U, w, VH, cutoff, cutoff_mode, max_bond, absorb, renorm
        )


class Z2FermionicArrayFlat(FermionicArrayFlat):
    static_symmetry = get_symmetry("Z2")



"""Abelian symmetric and fermionic Hamiltonians in terms of local operators."""


def make_edge_factory(t):
    """Ensure `t` is a function that takes two sites and returns an edge coeff."""
    if isinstance(t, dict):

        def edge_factory(cooa, coob):
            try:
                return t[(cooa, coob)]
            except KeyError:
                return t[(coob, cooa)]

    elif callable(t):
        edge_factory = t

    else:

        def edge_factory(cooa, coob):
            return t

    return edge_factory


def make_node_factory(U):
    """Ensure `U` is a function that takes a site and returns a node coeff."""
    if isinstance(U, dict):

        def node_factory(coo):
            return U[coo]

    elif callable(U):
        node_factory = U

    else:

        def node_factory(coo):
            return U

    return node_factory


def tfim_local_array(symmetry, jx=-1.0, hz=-3.0, coordinations=(1, 1)):
    """Build an abelian symmetric local operator for the transverse field
    Ising model::

        H = jx * sum_ij X_i X_j + hz * sum_i Z_i

    Note that its rotated into the x-basis so that the Z2 symmetry is manifest.
    """
    import quimb as qu

    from .utils import from_dense

    if symmetry != "Z2":
        raise ValueError(f"Symmetry {symmetry} not supported for TFIM.")

    index_map = [0, 1]

    try:
        ha, hb = hz
    except TypeError:
        ha = hb = hz

    I, Z, X = (qu.pauli(s, dtype="float64") for s in "IZX")

    h2 = (
        jx * (X & X)
        + (ha / coordinations[0]) * (Z & I)
        + (hb / coordinations[1]) * (I & Z)
    )

    return from_dense(
        h2.reshape(2, 2, 2, 2),
        symmetry=symmetry,
        index_maps=[index_map] * 4,
        duals=[False, False, True, True],
    )


def ham_tfim_from_edges(
    symmetry,
    edges,
    jx=-1.0,
    hz=-3.0,
):
    """Return a dict of local 2-body Hamiltonian abelian symmetric terms for
    the transverse field Ising model on the given lattice defined by `edges`::

        H = jx * sum_ij X_i X_j + hz * sum_i Z_i

    Note that its rotated into the x-basis so that the Z2 symmetry is manifest.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Should be "Z2".
    edges : Sequence[tuple[Hashable, Hashable]]
        A list of edges representing the lattice, each edge is a tuple of two
        nodes, each node is some hashable label.
    jx : float, dict, or callable, optional
        The coupling strength for the X-X interactions, by default -1.0. If a
        dict is given it should map edges to values, if a callable it should
        take the two sites as input.
    hz : float, dict or callable, optional
        The coupling strength for the Z interactions, by default -3.0. If a
        dict is given it should map sites to values, if a callable it should
        take the site as input.

    Returns
    -------
    dict[tuple[Hashable, Hashable], AbelianArray]
        A dictionary mapping edges to local Hamiltonian terms.
    """
    coordinations = {}
    for cooa, coob in edges:
        coordinations[cooa] = coordinations.setdefault(cooa, 0) + 1
        coordinations[coob] = coordinations.setdefault(coob, 0) + 1

    jx_factory = make_edge_factory(jx)
    hz_factory = make_node_factory(hz)

    return {
        (cooa, coob): tfim_local_array(
            symmetry,
            jx=jx_factory(cooa, coob),
            hz=(hz_factory(cooa), hz_factory(coob)),
            coordinations=(coordinations[cooa], coordinations[coob]),
        )
        for cooa, coob in edges
    }


def ham_heisenberg_from_edges(
    symmetry,
    edges,
    **kwargs,
):
    import quimb as qu

    from .utils import from_dense

    h2 = qu.ham_heis(2, **kwargs)
    index_map = [0, 1]

    h2 = from_dense(
        h2.reshape(2, 2, 2, 2),
        symmetry=symmetry,
        index_maps=[index_map] * 4,
        duals=[False, False, True, True],
    )

    return {(a, b): h2 for a, b in edges}


def ham_fermi_hubbard_from_edges(
    symmetry,
    edges,
    t=1.0,
    U=8.0,
    mu=0.0,
    like="numpy",
):
    """Return a dict of local 2-body Hamiltonian terms for the Fermi-Hubbard
    model on the given lattice defined by `edges`. The indices are ordered as
    (a, b, a', b'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively and similar for site b.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    edges : Sequence[tuple[Hashable, Hashable]]
        A list of edges representing the lattice, each edge is a tuple of two
        nodes, each node is some hashable label.
    t : float, dict or callable, optional
        The hopping parameter, by default 1.0. If a dict is given it should map
        edges to values, if a callable it should take the two sites as input.
    U : float dict or callable, optional
        The interaction parameter, by default 8.0. If a dict is given it should
        map sites to values, if a callable it should take the site as input.
    mu : float, optional
        The chemical potential, by default 0.0.
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    dict[tuple[Hashable, Hashable], FermionicArray]
        A dictionary mapping edges to local Hamiltonian terms.
    """
    from .fermionic_local_operators import fermi_hubbard_local_array

    coordinations = {}
    for cooa, coob in edges:
        coordinations[cooa] = coordinations.setdefault(cooa, 0) + 1
        coordinations[coob] = coordinations.setdefault(coob, 0) + 1

    t_factory = make_edge_factory(t)
    U_factory = make_node_factory(U)
    mu_factory = make_node_factory(mu)

    return {
        (cooa, coob): fermi_hubbard_local_array(
            symmetry,
            t=t_factory(cooa, coob),
            U=(U_factory(cooa), U_factory(coob)),
            mu=(mu_factory(cooa), mu_factory(coob)),
            coordinations=(coordinations[cooa], coordinations[coob]),
            like=like,
        )
        for cooa, coob in edges
    }


def ham_fermi_hubbard_spinless_from_edges(
    symmetry,
    edges,
    t=1.0,
    V=0.0,
    mu=0.0,
    like="numpy",
):
    """Return a dict of local 2-body Hamiltonian terms for the 'spinless
    Fermi-Hubbard' or ('t-V') model on the given lattice defined by `edges`.
    model on the given lattice defined by `edges`.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2" or "U1".
    edges : Sequence[tuple[Hashable, Hashable]]
        A list of edges representing the lattice, each edge is a tuple of two
        nodes, each node is some hashable label.
    t : float, optional
        The hopping parameter, by default 1.0.
    V : float, optional
        The nearest neighbor interaction parameter, by default 0.0.
    mu : float, optional
        The chemical potential, by default 0.0.
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    dict[tuple[Hashable, Hashable], FermionicArray]
        A dictionary mapping edges to local Hamiltonian terms.
    """
    from .fermionic_local_operators import fermi_hubbard_spinless_local_array

    coordinations = {}
    for cooa, coob in edges:
        coordinations[cooa] = coordinations.setdefault(cooa, 0) + 1
        coordinations[coob] = coordinations.setdefault(coob, 0) + 1

    t_factory = make_edge_factory(t)
    V_factory = make_edge_factory(V)
    mu_factory = make_node_factory(mu)

    return {
        (cooa, coob): fermi_hubbard_spinless_local_array(
            symmetry,
            t=t_factory(cooa, coob),
            V=V_factory(cooa, coob),
            mu=(mu_factory(cooa), mu_factory(coob)),
            coordinations=(coordinations[cooa], coordinations[coob]),
            like=like,
        )
        for cooa, coob in edges
    }



"""Abelian symmetric arrays with block sparse backend."""

from ..abelian_common import AbelianCommon
from ..common import SymmrayCommon
from ..symmetries import get_symmetry
from ..utils import DEBUG
from .sparse_array_common import SparseArrayCommon
from .sparse_data_common import BlockCommon
from .sparse_vector import BlockVector


class AbelianArray(
    SparseArrayCommon,
    BlockCommon,
    AbelianCommon,
    SymmrayCommon,
):
    """A block sparse array with symmetry constraints.

    Parameters
    ----------
    indices : tuple[BlockIndex]
        The indices of the array.
    charge : hashable, optionals
        The total charge of the array, if not given it will be inferred from
        either the first sector or set to the identity charge, if no sectors
        are given.
    blocks : dict[tuple[hashable], array_like]
        A mapping of each 'sector' (tuple of charges) to the data array.
    symmetry : str or Symmetry, optional
        The symmetry of the array, if not using a specific symmetry class.
    """

    __slots__ = (
        "_blocks",
        "_charge",
        "_indices",
        "_label",
        "_symmetry",
    )
    fermionic = False
    static_symmetry = None

    def __init__(
        self,
        indices,
        charge=None,
        blocks=(),
        symmetry=None,
        label=None,
    ):
        self._init_abelian(
            indices=indices,
            charge=charge,
            blocks=blocks,
            symmetry=symmetry,
            label=label,
        )

        if DEBUG:
            self.check()

    def new_with(self, indices, charge, blocks):
        """Create a new block sparse abelian array of the same class as this
        one. Unlike `copy`, this does not copy over any existing data and drops
        for example `label`.
        """
        return self._new_with_abelian(
            indices=indices,
            charge=charge,
            blocks=blocks,
        )

    def copy(self):
        """Copy this abelian block sparse array."""
        return self._copy_abelian()

    def copy_with(self, indices=None, charge=None, blocks=None):
        """A copy of this block array with some attributes replaced. Note that
        checks are not performed on the new properties, this is intended for
        internal use.
        """
        new = self._copy_with_abelian(
            indices=indices,
            charge=charge,
            blocks=blocks,
        )
        if DEBUG:
            new.check()
        return new

    def modify(self, indices=None, charge=None, blocks=None):
        """Modify this block array in place with some attributes replaced. Note
        that checks are not performed on the new properties, this is intended
        for internal use.
        """
        self._modify_abelian(indices=indices, charge=charge, blocks=blocks)
        if DEBUG:
            self.check()
        return self

    def _binary_blockwise_op(self, other, fn, missing=None, inplace=False):
        return self._binary_blockwise_op_abelian(
            other, fn, missing=missing, inplace=inplace
        )

    def _map_blocks(self, fn_block=None, fn_sector=None, fn_filter=None):
        """Map the blocks and their keys (sectors) of the array inplace."""
        self._map_blocks_blockcommon(fn_block, fn_sector, fn_filter)

    def transpose(self, axes=None, inplace=False):
        """Transpose this block sparse abelian array.

        Parameters
        ----------
        axes : tuple[int, ...] | None, optional
            A permutation of the axes to transpose the array by. If None,
            the axes will be reversed.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArray
        """
        return self._transpose_abelian(axes=axes, inplace=inplace)

    def conj(self, inplace=False) -> "AbelianArray":
        """Return the complex conjugate of this block array, including the
        indices."""
        return self._conj_abelian(inplace=inplace)

    def dagger(self, inplace=False) -> "AbelianArray":
        """Return the adjoint of this abelian array, including the
        indices and any subindex fusing information.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArray
        """
        return self._dagger_abelian(inplace=inplace)

    def _fuse_core(self, *axes_groups, mode="auto", inplace=False):
        return self._fuse_core_abelian(
            *axes_groups, mode=mode, inplace=inplace
        )

    def unfuse(self, axis, inplace=False) -> "AbelianArray":
        """Unfuse the ``axis`` index, which must carry subindex information,
        likely generated automatically from a fusing operation.

        Parameters
        ----------
        axis : int
            The axis to unfuse. It must have subindex information (`.subinfo`).
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArray
        """
        return self._unfuse_abelian(axis, inplace=inplace)

    def squeeze(self, axis=None, inplace=False):
        """Squeeze the block array, removing axes of size 1.

        Parameters
        ----------
        axis : int or tuple[int], optional
            The axis or axes to squeeze. If None, all axes of size 1 will be
            removed.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArray
        """
        return self._squeeze_abelian(axis=axis, inplace=inplace)

    def einsum(self, eq, preserve_array=False):
        """Einsum for abelian arrays, currently only single term.

        Parameters
        ----------
        eq : str
            The einsum equation, e.g. "abcb->ca". The output indices must be
            specified and only trace and permutations are allowed.
        preserve_array : bool, optional
            If tracing to a scalar, whether to return an AbelainArray object
            with no indices, or simply scalar itself (the default).

        Returns
        -------
        AbelianArray or scalar
        """
        return self._einsum_abelian(eq, preserve_array=preserve_array)

    def tensordot(self, other, axes=2, mode="auto", preserve_array=False):
        """Tensordot between two block sparse abelian symmetric arrays.

        Parameters
        ----------
        a, b : SparseArrayCommon
            The arrays to be contracted.
        axes : int or tuple[int]
            The axes to contract. If an integer, the last ``axes`` axes of
            ``a`` will be contracted with the first ``axes`` axes of ``b``. If
            a tuple, the axes to contract in ``a`` and ``b`` respectively.
        mode : {"auto", "fused", "blockwise"}
            The mode to use for the contraction. If "auto", it will choose
            between "fused" and "blockwise" based on the number of axes to
            contract.
        preserve_array : bool, optional
            Whether to return a scalar if the result is a scalar.
        """
        return self._tensordot_abelian(
            other,
            axes=axes,
            mode=mode,
            preserve_array=preserve_array,
        )

    def __matmul__(self, other, preserve_array=False):
        return self._matmul_abelian(other, preserve_array=preserve_array)

    def trace(self):
        """Compute the trace of the block array, assuming it is a square
        matrix.
        """
        return self._trace_abelian()

    def to_dense(self):
        return self._to_dense_abelian()

    def allclose(self, other, **allclose_opts):
        """Test whether this `AbelianArray` is close to another, that is,
        has all the same sectors, and the corresponding arrays are close.

        Parameters
        ----------
        other : AbelianArray
            The other array to compare to.
        allclose_opts
            Keyword arguments to pass to `allclose`.

        Returns
        -------
        bool
        """
        return self._allclose_abelian(other, **allclose_opts)

    def test_allclose(self, other, **allclose_opts):
        """Assert that this ``SparseArrayCommon`` is close to another,
        that is, has all the same sectors, and the corresponding arrays are
        close. Unlike `allclose`, this raises an AssertionError with details
        if not.

        Parameters
        ----------
        other : SparseArrayCommon
            The other array to compare to.
        allclose_opts
            Keyword arguments to pass to `allclose`.

        Raises
        ------
        AssertionError
            If the arrays are not close.
        """
        return self._test_allclose_abelian(other, **allclose_opts)

    # --------------------------- linalg methods ---------------------------- #

    def qr(
        self, stabilized=False, **kwargs
    ) -> tuple["AbelianArray", "AbelianArray"]:
        """QR decomposition of an AbelianArray.

        Parameters
        ----------
        x : AbelianArray
            The block symmetric array to decompose.
        stabilized : bool, optional
            Whether to use a stabilized QR decomposition, that is, with
            positive diagonal elements in the R factor. Default is False.

        Returns
        -------
        q : AbelianArray
            The orthogonal matrix.
        r : AbelianArray
            The upper triangular matrix.
        """
        return self._qr_abelian(stabilized=stabilized, **kwargs)

    def svd(
        self, **kwargs
    ) -> tuple["AbelianArray", "BlockVector", "AbelianArray"]:
        """SVD decomposition of an AbelianArray.

        Parameters
        ----------
        x : AbelianArray
            The block symmetric array to decompose.

        Returns
        -------
        u : AbelianArray
            The left singular vectors.
        s : BlockVector
            The singular values as a vector.
        vh : AbelianArray
            The right singular vectors (hermitian conjugated).
        """
        return self._svd_abelian(**kwargs)

    def eigh(self, **kwargs) -> tuple["BlockVector", "AbelianArray"]:
        """Eigenvalue decomposition of this assumed Hermitian AbelianArray.

        Parameters
        ----------
        x : AbelianArray
            The block symmetric array to decompose.

        Returns
        -------
        w : BlockVector
            The eigenvalues as a vector.
        u : AbelianArray
            The array of eigenvectors.
        """
        return self._eigh_abelian(**kwargs)

    def eigh_truncated(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        positive=0,
        **kwargs,
    ) -> tuple["AbelianArray", "BlockVector", "AbelianArray"]:
        """Truncated hermitian eigen-decomposition of this assumed hermitian
        block sparse abelian array.

        Parameters
        ----------
        cutoff : float, optional
            Absolute eigenvalue cutoff threshold.
        cutoff_mode : int or str, optional
            How to perform the truncation:

            - 1 or 'abs': trim values below ``cutoff``
            - 2 or 'rel': trim values below ``s[0] * cutoff``
            - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

        max_bond : int
            An explicit maximum bond dimension, use -1 for none.
        absorb : {-1, 0, 1, None}
            How to absorb the eigenvalues.

            - -1 or 'left': absorb into the left factor (U).
            - 0 or 'both': absorb the square root into both factors.
            - 1 or 'right': absorb into the right factor (VH).
            - None: do not absorb, return eigenvalues as a BlockVector.

        renorm : {0, 1}
            Whether to renormalize the eigenvalues (depends on `cutoff_mode`).

        Returns
        -------
        u : AbelianArray
            The abelian array of left eigenvectors.
        w : VectorCommon or None
            The vector of eigenvalues, or None if absorbed.
        uh : AbelianArray
            The abelian array of right eigenvectors.
        """
        return self._eigh_truncated_abelian(
            cutoff=cutoff,
            cutoff_mode=cutoff_mode,
            max_bond=max_bond,
            absorb=absorb,
            renorm=renorm,
            positive=positive,
            **kwargs,
        )

    def solve(self, b: "AbelianArray", **kwargs) -> "AbelianArray":
        """Solve the linear system `A @ x == b` for x, where A is this array.

        Parameters
        ----------
        b : AbelianArray
            The right-hand side array.

        Returns
        -------
        x : AbelianArray
            The solution array.
        """
        return self._solve_abelian(b, **kwargs)


# --------------------------------------------------------------------------- #


class Z2Array(AbelianArray):
    """A block array with Z2 symmetry."""

    static_symmetry = get_symmetry("Z2")

    def to_pyblock3(self, flat=False):
        from pyblock3.algebra.core import SparseTensor, SubTensor
        from pyblock3.algebra.fermion_symmetry import Z2

        blocks = [
            SubTensor(array, q_labels=tuple(map(Z2, sector)))
            for sector, array in self.get_sector_block_pairs()
        ]

        data = SparseTensor(blocks)

        if flat:
            from pyblock3.algebra.flat import FlatSparseTensor

            data = FlatSparseTensor.from_sparse(data)

        data.shape = self.shape

        return data


class U1Array(AbelianArray):
    """A block array with U1 symmetry."""

    static_symmetry = get_symmetry("U1")

    def to_pyblock3(self, flat=False):
        from pyblock3.algebra.core import SparseTensor, SubTensor
        from pyblock3.algebra.fermion_symmetry import U1

        blocks = [
            SubTensor(array, q_labels=tuple(map(U1, sector)))
            for sector, array in self.get_sector_block_pairs()
        ]

        data = SparseTensor(blocks)

        if flat:
            from pyblock3.algebra.flat import FlatSparseTensor

            data = FlatSparseTensor.from_sparse(data)

        data.shape = self.shape

        return data

    def to_yastn(self, **config_opts):
        import yastn

        t = yastn.Tensor(
            config=yastn.make_config(sym="U1", **config_opts),
            s=tuple(-1 if ix.dual else 1 for ix in self.indices),
            n=self.charge,
        )
        for sector, array in self.get_sector_block_pairs():
            t.set_block(ts=sector, Ds=array.shape, val=array)

        return t


class Z2Z2Array(AbelianArray):
    """A block array with Z2 x Z2 symmetry."""

    static_symmetry = get_symmetry("Z2Z2")


class U1U1Array(AbelianArray):
    """A block array with U1 x U1 symmetry."""

    static_symmetry = get_symmetry("U1U1")



"""Index objects for abelian arrays with block sparse backend."""

import numbers
import warnings

from ..index_common import Index, SubInfo
from ..utils import hasher


class BlockIndex(Index):
    """An index of a block sparse, abelian symmetric tensor. This is intended
    to be used immutably.

    Parameters
    ----------
    chargemap : dict[int, int]
        A mapping from charge to size.
    dual : bool, optional
        Whether the index is 'dual' or not, i.e. whether the flow is
        'outwards' / (+ve) / ket-like = ``False`` or 'inwards' / (-ve) /
        bra-like = ``True``. The sign of charge contributions is then
        ``(-1) ** dual``.
    subinfo : SubIndexInfo, optional
        Information about the subindices of this index and their extents if
        this index was formed from fusing.
    linearmap : Sequence[tuple[hashable, int], ...], optional
        If provided, a sequence of (charge, offset) pairs for a linear ordering
        of this index. This can be supplied instead of ``chargemap``, in which
        case the ``chargemap`` is built from this. If not provided, it is built
        on demand from ``chargemap``, assuming blocks of sorted charges.
    """

    __slots__ = (
        "_chargemap",
        "_dual",
        "_subinfo",
        "_hashkey",
        "_linearmap",
    )

    def __init__(
        self,
        chargemap=None,
        dual=False,
        subinfo=None,
        linearmap=None,
    ):
        if chargemap is None:
            if linearmap is None:
                raise ValueError("Must provide either chargemap or linearmap.")
            # build chargemap from linearmap
            chargemap = {}
            for c, _ in linearmap:
                chargemap[c] = chargemap.get(c, 0) + 1

        # ensure always sorted
        if not isinstance(chargemap, dict):
            self._chargemap = dict(sorted(chargemap))
        else:
            self._chargemap = dict(sorted(chargemap.items()))
        self._dual = bool(dual)
        self._subinfo = subinfo
        self._hashkey = None
        self._linearmap = linearmap

    def copy_with(self, **kwargs):
        """A copy of this index with some attributes replaced. Note that checks
        are not performed on the new propoerties, this is intended for internal
        use.
        """
        new = self.__new__(self.__class__)
        keep_linearmap = True

        if "chargemap" in kwargs:
            chargemap = kwargs.pop("chargemap")
            new._chargemap = (
                dict(sorted(chargemap.items()))
                if isinstance(chargemap, dict)
                else dict(sorted(chargemap))
            )
            keep_linearmap = False
        else:
            new._chargemap = self._chargemap.copy()

        if "dual" in kwargs:
            new._dual = bool(kwargs.pop("dual"))
        else:
            new._dual = self._dual

        # need to pop from kwargs to handle 'not-set' vs 'set-to-None'
        if "subinfo" in kwargs:
            new._subinfo = kwargs.pop("subinfo")
        else:
            new._subinfo = self._subinfo

        # only keep linearmap if chargemap unchanged
        if keep_linearmap:
            new._linearmap = self._linearmap
        else:
            new._linearmap = None

        # always recompute this
        new._hashkey = None

        if kwargs:
            raise TypeError(f"Unexpected keyword arguments: {kwargs}")

        return new

    def to_flat(self):
        """Convert this index to a flat index."""
        from ..flat.flat_index import FlatIndex

        charges = sorted(self.charges)
        num_charges = len(charges)
        charge_size = max(self.sizes)

        if (num_charges != 1) and (charges != list(range(num_charges))):
            raise ValueError(
                "FlatIndex requires a single charge or 0...N-1 as charges."
            )

        if self.subinfo is not None:
            warnings.warn(
                "Converting subinfo to flat not supported yet, dropping it."
            )

        return FlatIndex(
            num_charges=num_charges,
            charge_size=charge_size,
            dual=self.dual,
            linearmap=self._linearmap,
        )

    @property
    def chargemap(self):
        """A mapping from charge to size."""
        return self._chargemap

    @property
    def linearmap(self):
        """A sequence of (charge, offset) pairs for each element of this index,
        in linear order. This is built on demand if not provided at creation.
        """
        if self._linearmap is None:
            # build a default linearmap in sorted blocks
            linearmap = []
            for c, d in sorted(self._chargemap.items()):
                linearmap.extend((c, i) for i in range(d))
            self._linearmap = tuple(linearmap)
        return self._linearmap

    @property
    def subshape(self):
        if self._subinfo is None:
            return None
        return self._subinfo.subshape

    @property
    def charges(self):
        """The charges of this index."""
        return self._chargemap.keys()

    @property
    def sizes(self):
        """The sizes of the blocks of this index."""
        return self._chargemap.values()

    @property
    def size_total(self):
        """The total size of this index, i.e. the sum of the sizes of all
        blocks.
        """
        return sum(self._chargemap.values())

    @property
    def num_charges(self):
        """The number of charges."""
        return len(self._chargemap)

    def conj(self):
        """A copy of this index with the dualness reversed."""
        dual = not self.dual
        subinfo = None if self.subinfo is None else self.subinfo.conj()
        return self.copy_with(dual=dual, subinfo=subinfo)

    def drop_charges(self, charges):
        """Return a new index with all charges in ``charges`` removed.

        Parameters
        ----------
        charges : Sequence[hashable]
            The charges to remove.

        Returns
        -------
        BlockIndex
            A new index with the charges removed.
        """
        return self.copy_with(
            chargemap={
                c: d for c, d in self._chargemap.items() if c not in charges
            },
            subinfo=(
                None
                if self.subinfo is None
                else self.subinfo.drop_charges(charges)
            ),
        )

    def select_charge(self, charge, subselect=None) -> "BlockIndex":
        """Drop all but the specified charge from this index.

        Parameters
        ----------
        charge : hashable
            The charge to keep.
        subselect : slice or array_like, optional
            If provided, a range of indices within the selected charge block
            to keep. If not provided, the entire block is kept.

        Returns
        -------
        BlockIndex
        """
        drop = set(self._chargemap)
        drop.remove(charge)
        new = self.drop_charges(drop)
        if subselect is None:
            return new

        # need to update charge size as well
        new_chargemap = new._chargemap.copy()
        current_size = new_chargemap[charge]

        if isinstance(subselect, slice):
            start, stop, step = subselect.indices(current_size)
            size = len(range(start, stop, step))
            new_chargemap[charge] = size
        elif hasattr(subselect, "size"):  # numpy array or similar
            new_chargemap[charge] = subselect.size
        else:
            new_chargemap[charge] = len(subselect)

        # XXX: could we update subinfo rather than remove?
        return new.copy_with(chargemap=new_chargemap, subinfo=None)

    def size_of(self, c):
        """The size of the block with charge ``c``."""
        return self._chargemap[c]

    def linear_to_charge_and_offset(self, i):
        """Given a linear index ``i`` into this index (as if it were a dense
        array), return the corresponding charge and offset within that charge
        block.

        Parameters
        ----------
        i : int
            The linear index into this index.

        Returns
        -------
        charge : hashable
            The charge corresponding to the linear index.
        offset : int
            The offset within the charge block corresponding to the linear
            index.
        """
        return self.linearmap[i]

    def check(self):
        """Check that the index is well-formed, i.e. all sizes are positive."""
        for c, d in self._chargemap.items():
            if d <= 0:
                raise ValueError(
                    f"Size of charge {c} is {d}, must be positive."
                )
            if not isinstance(d, numbers.Integral):
                raise ValueError(f"Size of charge {c} is {d}, must be an int.")

        assert sorted(self._chargemap) == list(self._chargemap)

        if self.subinfo:
            assert self.size_total == sum(
                d
                for extent in self.subinfo.extents.values()
                for d in extent.values()
            )

        if self._linearmap is not None:
            if len(self._linearmap) != self.size_total:
                raise ValueError(
                    f"Index map length {len(self._linearmap)} does not "
                    f"match index size {self.size_total}."
                )
            seen = set()
            for c, i in self._linearmap:
                if c not in self._chargemap:
                    raise ValueError(
                        f"Index map charge {c} not in chargemap "
                        f"{self._chargemap}."
                    )
                if i < 0 or i >= self._chargemap[c]:
                    raise ValueError(
                        f"Index map offset {i} out of bounds for charge {c} "
                        f"with size {self._chargemap[c]}."
                    )
                if (c, i) in seen:
                    raise ValueError(
                        f"Index map has duplicate entry {(c, i)}."
                    )
                seen.add((c, i))

    def matches(self, other):
        """Whether this index matches ``other`` index, namely, whether the
        ``chargemap`` of each matches, their dualnesses are opposite, and also
        whether their subindices match, if they have any. For debugging.

        Parameters
        ----------
        other : BlockIndex
            The other index to compare to.
        """
        return (
            dicts_dont_conflict(self._chargemap, other._chargemap)
            and (self.dual ^ other.dual)
            and (
                (self.subinfo is other.subinfo is None)
                or (self.subinfo.matches(other.subinfo))
            )
        )

    def hashkey(self):
        """Get a hash key for this index."""
        if getattr(self, "_hashkey", None) is None:
            self._hashkey = hasher(
                (
                    tuple(self._chargemap.items()),
                    self._dual,
                    self._subinfo.hashkey() if self._subinfo else None,
                )
            )
        return self._hashkey

    def __hash__(self):
        # disable as hash(-1) == hash(-2) causes problems
        # XXX: use hashkey?
        raise NotImplementedError

    def __str__(self):
        lines = [
            f"({self.size_total} = "
            f"{'+'.join(map(str, self._chargemap.values()))} "
            f": {'-' if self.dual else '+'}"
            f"[{','.join(map(str, self._chargemap.keys()))}])"
        ]

        if self.subinfo:
            for charge, extent in sorted(self.subinfo.extents.items()):
                subcharges = extent.keys()
                subsizes = extent.values()
                lines.append(
                    f"    {charge} ; "
                    f"({'+'.join(map(str, subsizes))}) : "
                    f"[{','.join(map(str, subcharges))}]"
                )

        return "\n".join(lines)

    def __repr__(self):
        return "".join(
            [
                f"{self.__class__.__name__}(",
                f"chargemap={self._chargemap}, dual={self.dual}",
                (
                    f", subinfo={self.subinfo}"
                    if self.subinfo is not None
                    else ""
                ),
                ")",
            ]
        )


def dicts_dont_conflict(da, db):
    """Check that two dictionaries don't conflict, i.e. they have no keys in
    common with different values."""
    for k, va in da.items():
        vb = db.get(k, None)
        if vb is not None and va != vb:
            return False
    return True


class SubIndexInfo(SubInfo):
    """Holder class for storing the relevant information for unfusing.

    Parameters
    ----------
    indices : tuple[BlockIndex]
        The indices (ordered) that were fused to make this index.
    extents : dict[hashable, dict[hashable, int]]
        A mapping of each charge of the fused index to a mapping of each
        subsector (combination of sub charges) to the size of that subsector.
        This should not be mutated after creation.
    """

    __slots__ = ("_extents", "_indices", "_hashkey")

    def __init__(self, indices, extents):
        self._indices = indices
        self._extents = extents
        self._hashkey = None

    @property
    def extents(self):
        """A mapping of each charge of the fused index to a mapping of each
        subsector (combination of sub charges) to the size of that subsector.
        """
        return self._extents

    @property
    def subshape(self):
        return tuple(ix.size_total for ix in self._indices)

    def copy_with(self, indices=None, extents=None):
        """A copy of this subindex information with some attributes replaced.
        Note that checks are not performed on the new properties, this is
        intended for internal use.
        """
        new = self.__new__(self.__class__)
        new._indices = self._indices if indices is None else indices
        new._extents = self._extents if extents is None else extents
        new._hashkey = None
        return new

    def conj(self):
        """A copy of this subindex information with the relevant dualnesses
        reversed.
        """
        return self.copy_with(indices=tuple(ix.conj() for ix in self._indices))

    def drop_charges(self, charges):
        """Get a copy of this subindex information with the charges in
        ``charges`` discarded.
        """
        return self.copy_with(
            extents={
                c: extent
                for c, extent in self._extents.items()
                if c not in charges
            },
        )

    def matches(self, other):
        """Whether this subindex information matches ``other`` subindex
        information, namely, whether the ``indices`` and ``extents`` match.
        For debugging.
        """
        return all(
            i.matches(j) for i, j in zip(self._indices, other._indices)
        ) and dicts_dont_conflict(self._extents, other._extents)

    def hashkey(self):
        """Get a string hash key for this subindex information. This is cached
        after the first call.
        """
        if getattr(self, "_hashkey", None) is None:
            self._hashkey = hasher(
                (
                    tuple(ix.hashkey for ix in self._indices),
                    tuple(
                        (c, tuple(extent.items()))
                        for c, extent in self._extents.items()
                    ),
                )
            )
        return self._hashkey

    def __hash__(self):
        # disable as hash(-1) == hash(-2) causes problems
        # XXX: use hashkey?
        raise NotImplementedError

    def __repr__(self):
        return "".join(
            [
                f"{self.__class__.__name__}(",
                f"indices={self._indices}, ",
                f"extents={self._extents}",
                ")",
            ]
        )



"""Methods that apply to abelian arrays with flat backend, both
fermionic and bosonic.
"""

import functools
import math
import numbers
import operator
from itertools import repeat

import autoray as ar
import cotengra as ctg

from ..sparse.sparse_abelian_array import AbelianArray
from ..sparse.sparse_array_common import (
    calc_fuse_group_info,
    parse_tensordot_axes,
)
from ..utils import DEBUG, get_array_cls
from .flat_index import FlatIndex, FlatSubIndexInfo
from .flat_vector import FlatVector

try:
    from einops import rearrange as _einops_rearrange
    from einops import repeat as _einops_repeat
    from einops.array_api import rearrange as _einops_rearrange_api
    from einops.array_api import repeat as _einops_repeat_api

    # want to support both standard and array_api versions

    def einops_rearrange(tensor, *args, **kwargs):
        if hasattr(tensor, "__array_namespace__"):
            return _einops_rearrange_api(tensor, *args, **kwargs)
        return _einops_rearrange(tensor, *args, **kwargs)

    def einops_repeat(tensor, *args, **kwargs):
        if hasattr(tensor, "__array_namespace__"):
            return _einops_repeat_api(tensor, *args, **kwargs)
        return _einops_repeat(tensor, *args, **kwargs)

except ImportError:

    def missinglib(*args, name, **kwargs):
        raise ImportError(f"'{name}' required for this function.")

    einops_rearrange = functools.partial(missinglib, name="einops.rearrange")
    einops_repeat = functools.partial(missinglib, name="einops.repeat")


def lexsort_sectors(sectors, stable=True):
    """Given a sequence of columns of positive integers, or equivalently a
    matrix of shape (num_sectors, num_charges), find the indices that
    sort them lexicographically, such that the first column is the most
    significant, then the second, and so forth.

    Parameters
    ----------
    sectors : array_like or sequence[array_like]
        The sectors to sort, each column should be a 1D array of positive
        integer charges. Either supplied as a 2D array, or a sequence of 1D
        arrays (columns of charges), in which case they will be stacked along a
        second axis.
    stable : bool, optional
        Whether to use a stable sort. Default is True, which uses the
        `argsort` function with the `stable` parameter set to True. If False,
        it will use the default sorting behavior which may not be stable.

    Returns
    -------
    array_like
        The indices that would sort the stack of sectors lexicographically.

    Examples
    --------

        >>> sectors = np.array([[4, 1, 0], [3, 2, 1], [3, 1, 0], [2, 0, 1]])
        >>> k = lexsort_sectors(sectors)
        >>> sectors[k]
        array([[2, 0, 1],
               [3, 1, 0],
               [3, 2, 1],
               [4, 1, 0]])
    """
    # XXX: this is a bottleneck

    if ar.is_array(sectors):
        ncol = sectors.shape[1]
        cols = tuple(sectors[:, c] for c in range(ncol))
    else:
        cols = sectors

    xp = ar.get_namespace(cols[0])

    if len(cols) == 1:
        return xp.argsort(cols[0], stable=stable)

    strides = [1]

    for col in cols[:0:-1]:
        strides.insert(0, (xp.max(col) + 1) * strides[0])

    ranks = functools.reduce(
        operator.add, (w * col for w, col in zip(strides, cols))
    )

    return xp.argsort(ranks, stable=stable)

    # # alternative implementation:

    # if not ar.is_array(sectors):
    #     sectors = ar.do("stack", sectors, axis=1)

    # backend = ar.infer_backend(sectors)

    # n = ar.do("shape", sectors, like=backend)[1]
    # limits = ar.do("max", sectors, axis=0, like=backend) + 1

    # # ws = ar.do("ones", (1, n), dtype=sectors.dtype, like=backend)

    # ws = [ar.do("array", 1, like=sectors)] * n
    # for ax in range(n - 2, -1, -1):
    #     # reverse cumulative product to get 'strides'
    #     ws[ax] = ws[ax + 1] * limits[ax]
    # ws = ar.do("stack", tuple(ws), like=sectors)
    # ws = ar.do("reshape", ws, (1, n))

    # ranks = ar.do("sum", ws * sectors, axis=1, like=backend)
    # kord2 = ar.do("argsort", ranks, stable=stable, like=backend)

    # assert ar.do("all", kord1 == kord2)
    # return kord1


def zn_combine(order, sectors, duals=None, like=None):
    """Implement vectorized addition modulo group order, with signature.

    Parameters
    ----------
    order : int
        The order of the symmetry group, i.e., the number of distinct charges
        in each axis. E.g. 2 corresponds to Z2 symmetry.
    sectors : array_like
        The stack of sectors, with shape (num_blocks, num_charges). Each
        row represents a subsector.
    duals : sequence[bool] | None, optional
        The dualness of each index, i.e., whether the charge contributes
        positively or negatively. If not given, it will be assumed that all
        charges are positive.
    like : str or array_like, optional
        The array-like object to use as a reference for the output type and
        backend. If not given, will be inferred..
    """
    if order == 2:
        # self inverse, no need to check duals
        return ar.do("sum", sectors, axis=-1, like=like) % 2

    if (duals is not None) and any(duals):
        # turn duals into phases
        signs = [dual * -2 + 1 for dual in duals]
        # broadcasted multiply
        signs = ar.do("array", signs, like=like)
        signs = ar.do("reshape", signs, (1, -1))
        signed_sectors = sectors * signs
    else:
        signed_sectors = sectors

    return ar.do("sum", signed_sectors, axis=-1, like=like) % order


def _calc_fused_sectors_subkeys_slice(
    num_groups,
    axes_groups,
    new_sectors,
    old_sectors,
    pos,
    ndim,
    new_ndim,
    order,
    group_singlets,
    backend,
):
    """Calculate new sectors and group subkeys, by slicing the existing
    sectors.
    """
    # now we compute subcharge information for each group
    # first we reshape the old sectors given the sort above:
    #     (*fused_charges, *unfused_charges, --locked dim, *subcharges)
    # whichever is the last charge of the first two groups is locked,
    # and so taken away from the reshaping.
    keys_reshaper = []
    axes_seen = 0
    g_locked_map = {}

    for g in range(num_groups):
        axes_seen += len(axes_groups[g])
        if axes_seen == ndim:
            # if all axes are being fused, the last one is locked
            keys_reshaper.append(1)
            if num_groups > 1:
                # all axes are fused, multiple new charges, last of which
                # is locked to prev, so we can't take a slice across it
                # directly, nor will it be sorted
                g_locked_map[g] = g - 1
        else:
            # one dimension for each fused charge
            keys_reshaper.append(order)
    # one dimension for all the unfused axes
    keys_reshaper.append(-1)
    # then one dimension for each *extra* subcharge in each group
    nsubaxes = 0
    for group in axes_groups:
        # each group has one locked charge within it
        nsub = len(group) - 1
        keys_reshaper.extend([order] * nsub)
        nsubaxes += nsub

    # reshape! including finally one dimension for the row of charges
    new_sectors = ar.do("reshape", new_sectors, (*keys_reshaper, new_ndim))
    old_sectors = ar.do("reshape", old_sectors, (*keys_reshaper, ndim))

    # drop sub charge axes from new_sectors
    if nsubaxes == 1:
        new_sectors = ar.do("take", new_sectors, 0, axis=num_groups + 1)
    elif nsubaxes > 1:
        new_sectors = new_sectors[
            (
                *repeat(slice(None), num_groups),  # fused charges
                slice(None),  # unfused axes
                *repeat(0, nsubaxes),  # sub charges, take first (arbitrary)
                slice(None),  # row of charges
            )
        ]

    # then we take slices across these to strore subcharge information
    # for unfusing
    subkeys = []
    for g in range(num_groups):
        if g in group_singlets:
            # no need to record subcharges
            subkeys.append(None)
            continue

        subkey_selector = []

        # is this group is locked, need to slice over previous group
        g_lock = g_locked_map.get(g, g)

        # fix other grouped charges to any value, take slice across g
        for go in range(num_groups):
            if go == g_lock:
                subkey_selector.append(slice(None))
            else:
                subkey_selector.append(0)

        # fix stack of unfused axes to any value
        subkey_selector.append(0)

        # then we fix the subcharges for other groups to any value
        # and slice across fused subcharges
        for go in range(num_groups):
            m = len(axes_groups[go]) - 1
            if go == g:
                subkey_selector.extend([slice(None)] * m)
            else:
                subkey_selector.extend([0] * m)

        # finally we take the slice across the legs
        # i.e. the row of charges making up the sector
        subkey_selector.append(slice(None))

        # take the slice!
        subkey_selector = tuple(subkey_selector)
        subkey = old_sectors[subkey_selector]
        # flatten into (overall_charge, subcharges, row)
        subkey = ar.do("reshape", subkey, (keys_reshaper[g_lock], -1, ndim))
        # just get axes relevant to this group
        subkey = subkey[:, :, axes_groups[g]]

        if g_lock != g:
            # this group axis is locked to the previous group,
            # so we need to do an additional sort the overall charge axis
            gcharges = new_sectors[
                (
                    *repeat(0, g - 1),  # other groups
                    slice(None),  # group we are locked to
                    0,  # locked g axis
                    0,  # unfused axes
                    pos + g,
                )
            ]
            kord = ar.do("argsort", gcharges, like=backend)
            subkey = subkey[kord]

        subkeys.append(subkey)

    # reflatten new keys into stack
    new_sectors = ar.do("reshape", new_sectors, (-1, new_ndim), like=backend)

    return new_sectors, subkeys


def _calc_fused_sectors_subkeys_create(
    unmerged_batch_sizes,
    num_groups,
    axes_groups,
    new_sectors,
    old_sectors,
    group_duals,
    ndim,
    order,
    duals,
    like,
):
    """Calculate new sectors and group subkeys, by explicit generation."""
    # then we update the sectors, these have been sorted already by
    # grouped charge, each group being of equal size / stride:
    stride = math.prod(unmerged_batch_sizes.values())
    new_sectors = new_sectors[::stride]

    if num_groups == 1 and len(axes_groups[0]) == ndim:
        # full fuse, only one overall charge, subkeys are all current keys
        # but we do need to take into account possible permutation
        subkeys0 = ar.do(
            "reshape", old_sectors[..., axes_groups[0]], (1, -1, ndim)
        )
        subkeys = [subkeys0]
    else:
        subkeys = [
            build_cyclic_keys_by_charge(
                ndim=len(gaxes),
                order=order,
                duals=[duals[ax] != group_duals[g] for ax in gaxes],
                like=like,
            )
            for g, gaxes in enumerate(axes_groups)
        ]

    return new_sectors, subkeys


def _calc_fuse_rearrange_pattern(
    num_groups,
    axes_groups,
    axes_before,
    axes_after,
    axes_ncharges,
    ndim,
):
    # now we create the unmerge/merge pattern for einops:
    # heres a full example for 2 groups and 6 axes:
    # axes_groups = ((5, 2), (4, 1))
    # '(B B0 B1) p0 p1 p2 p3 p4 p5 -> B p0 (B0 p5 p2) (B1 p4 p1) p3'

    # LHS, first we 'unfuse' the block index
    # with one new axis for each group, `ax0 -> (B B0 B1 B2 ...)`
    pattern = ["(B"]
    for g in range(num_groups):
        pattern.append(f" B{g}")
    pattern.append(")")

    # then we label each of the input axes `... p0 p1 p2 ...`
    for ax in range(ndim):
        pattern.append(f" p{ax}")

    # RHS, start with the new block index
    pattern.append(" -> B")

    # then add the unfused output axes before the groups
    for ax in axes_before:
        pattern.append(f" p{ax}")

    # then the groups, each looks like `(B0 p5 p2 ...)`, one dimension
    # coming from the batch index, and the rest real axis fusions
    unmerged_batch_sizes = {}
    for g, gaxes in enumerate(axes_groups):
        pattern.append(f" (B{g}")
        for ax in gaxes:
            pattern.append(f" p{ax}")

            # XXX: need to handle singlet dims somehow?
            # if axes_ncharges[ax] > 1:

            bax = f"B{g}"
            # keep track of the unmerged
            if bax in unmerged_batch_sizes:
                unmerged_batch_sizes[bax] *= axes_ncharges[ax]
            else:
                unmerged_batch_sizes[bax] = 1
        pattern.append(")")

    # then add the unfused output axes after the groups
    for ax in axes_after:
        pattern.append(f" p{ax}")

    pattern = "".join(pattern)
    return pattern, unmerged_batch_sizes


class FlatArrayCommon:
    def _init_abelian(
        self,
        sectors,
        blocks,
        indices,
        symmetry=None,
        label=None,
    ):
        self._init_flatcommon(sectors, blocks)
        self._symmetry = self.get_class_symmetry(symmetry)
        self._label = label
        self._indices = tuple(
            # allow sequence of duals to be supplied directly
            x
            if isinstance(x, FlatIndex)
            else FlatIndex(
                num_charges=1 if self.ndim == 1 else self.order,
                charge_size=d,
                dual=x,
            )
            for x, d in zip(indices, self.shape_block)
        )

    def _check_abelian(self):
        assert len(self._sectors) == len(self._blocks)
        assert self.ndim == len(self._indices)
        assert self.ndim == len(self._sectors[0])
        # check blocks all have the same overall charge
        sector_charges = ar.do(
            "unique", zn_combine(self.order, self._sectors, self.duals)
        )
        assert ar.do("size", sector_charges) == 1
        for ix, ds in zip(self._indices, self.shape_block):
            ix.check()
            assert ds == ix.charge_size

        if self._blocks.__class__.__name__ != "Placeholder":
            assert ar.do("all", ar.do("isfinite", self._blocks))

    def _new_with_abelian(self, sectors, blocks, indices):
        new = self._new_with_flatcommon(sectors, blocks)
        new._indices = indices
        new._symmetry = self._symmetry
        new._label = None
        return new

    def _copy_abelian(self, deep=False) -> "FlatArrayCommon":
        """Create a copy of the array."""
        new = self._copy_flatcommon(deep=deep)
        new._indices = self._indices
        new._symmetry = self._symmetry
        new._label = self._label
        return new

    def _copy_with_abelian(
        self,
        sectors=None,
        blocks=None,
        indices=None,
    ) -> "FlatArrayCommon":
        """A copy of this flat array with some attributes replaced. Note that
        checks are not performed on the new properties, this is intended for
        internal use.
        """
        new = self._copy_with_flatcommon(sectors=sectors, blocks=blocks)
        new._indices = self._indices if indices is None else indices
        new._symmetry = self._symmetry
        new._label = self._label
        return new

    def _modify_abelian(
        self,
        sectors=None,
        blocks=None,
        indices=None,
    ) -> "FlatArrayCommon":
        """Modify this flat array in place with some attributes replaced. Note
        that checks are not performed on the new properties, this is intended
        for internal use.
        """
        self._modify_flatcommon(sectors=sectors, blocks=blocks)
        if indices is not None:
            self._indices = indices
        return self

    def _set_params_abelian(self, params):
        """Set the underlying array blocks."""
        self._set_params_flatcommon(params)

        new_indices = []
        for ix in self._indices:
            if ix._linearmap is None:
                new_ix = ix
            else:
                try:
                    new_ix = ix.copy_with(
                        linearmap=ar.do("asarray", ix._linearmap, like=params)
                    )
                except ImportError:
                    # params is possibly a placeholder of some kind
                    # -> no backend.array function available
                    pass

            new_indices.append(new_ix)
        self._indices = tuple(new_indices)

    @property
    def order(self) -> int:
        """Get the order of the symmetry group."""
        return self._symmetry.N

    @property
    def label(self):
        """The label of the array, possibly used for ordering odd parity
        fermionic modes."""
        return self._label

    @property
    def charge(self):
        """Compute the overall charge of the array."""
        return zn_combine(
            self.order,
            self._sectors[[0], :],
            self.duals,
            like=self.backend,
        )[0]

    @property
    def shape(self) -> tuple[int, ...]:
        """Get the effective shape of the array."""
        return tuple(ix.size_total for ix in self._indices)

    @property
    def size(self) -> int:
        """Get the total size of the array, i.e., the product of all dimensions
        in the effective shape."""
        return functools.reduce(operator.mul, self.shape, 1)

    @classmethod
    def from_scalar(cls, x, symmetry=None) -> "FlatArrayCommon":
        """Create a flat abelian array from a scalar."""
        sectors = [[]]
        indices = ()
        blocks = ar.do("reshape", x, (1,))
        return cls(sectors, blocks, indices, symmetry=symmetry)

    def _to_blocksparse_abelian(self, **kwargs) -> AbelianArray:
        """Create a blocksparse abelian array from this flat abelian array."""
        cls = get_array_cls(
            self.symmetry,
            self.fermionic,
            False,
        )

        blocks = {}
        for i in range(self.num_blocks):
            sector = tuple(map(int, self._sectors[i]))
            block = self._blocks[i]
            blocks[sector] = block

        # XXX: convert indices directly to include e.g. linearmap information

        return cls.from_blocks(
            blocks,
            duals=self.duals,
            symmetry=self.symmetry,
            **kwargs,
        )

    def _to_dense_abelian(self):
        return self._to_blocksparse_abelian()._to_dense_abelian()

    def _map_blocks_abelian(self, fn_sector=None, fn_block=None):
        if fn_sector is not None:
            self._sectors = fn_sector(self._sectors)
        if fn_block is not None:
            self._blocks = fn_block(self._blocks)

    def get_sorting_indices(
        self,
        axes=None,
        all_axes=None,
    ):
        """Get the indices that would lexicgraphically sort the stack of
        sectors according to the values of charges in the specified axes,
        optionally filling in the rest of the axes with the remaining axes in
        the order they appear.

        Parameters
        ----------
        axes : int | tuple[int, ...], optional
            The axes to sort by. If a single integer is given, it will be
            interpreted as the axis to sort by. If a tuple of integers is
            given, it will be interpreted as the axes to sort by in order.
            Default is None, if all_axes is also None or True, this will sort
            all axes in their current order.
        all_axes : bool, optional
            Whether to include all non-specified axes as tie-breakers, after
            the specified axes. If ``None``, the default, this will be True
            if `axes` is not supplied explicitly, and False otherwise.
        """
        if axes is None:
            axes = ()
            if all_axes is None:
                all_axes = True
        elif all_axes is None:
            all_axes = False

        # ensure tuple and handle negative axes
        if not isinstance(axes, (tuple, list)):
            if axes < 0:
                axes = axes + self.ndim
            axes = (axes,)
        else:
            axes = tuple(ax if ax >= 0 else ax + self.ndim for ax in axes)

        if all_axes:
            # include all non-specified axes as tie-breakers
            axes = (*axes, *(ax for ax in range(self.ndim) if ax not in axes))

        cols = self._sectors[:, axes]
        return lexsort_sectors(cols)

    def _transpose_abelian(
        self,
        axes=None,
        inplace=False,
    ) -> "FlatArrayCommon":
        """Transpose this flat abelian array.

        Parameters
        ----------
        axes : tuple[int, ...] | None, optional
            A permutation of the axes to transpose the array by. If None,
            the axes will be reversed.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        FlatArrayCommon
        """
        if axes is None:
            # reverse the axes
            axes = tuple(range(self.ndim - 1, -1, -1))

        axes = tuple(map(int, axes))

        # transpose block as usual, but with broadcasted block axis
        new_blocks = ar.do(
            "transpose",
            self._blocks,
            (0, *(ax + 1 for ax in axes)),
            like=self.backend,
        )

        new_sectors = self._sectors[:, axes]
        new_indices = tuple(self._indices[ax] for ax in axes)

        return self._modify_or_copy(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
            inplace=inplace,
        )

    def _conj_abelian(self, inplace=False) -> "FlatArrayCommon":
        """Return the complex conjugate of this block array, including the
        indices and any subindex fusing information.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        FlatArrayCommon
        """
        new_sectors = self._sectors
        new_blocks = ar.do("conj", self._blocks, like=self.backend)
        new_indices = tuple(ix.conj() for ix in self._indices)
        return self._modify_or_copy(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
            inplace=inplace,
        )

    def expand_dims(
        self, axis, c=None, dual=None, inplace=False
    ) -> "FlatArrayCommon":
        """Expand the shape of an abelian array.

        Parameters
        ----------
        axis : int
            The position along which to expand.
        c : hashable, optional
            The charge to insert at the new axis. If not given, a zero charge
            will be inserted.
        dual : bool, optional
            The dual-ness of the new index. If not given, it will be inherited
            from the axis before or after, if any. If there is no axis before
            or after, it will default to `False`.
        inplace : bool, optional
            Whether to perform the operation inplace.

        Returns
        -------
        FlatArrayCommon
        """
        if axis < 0:
            axis += self.ndim + 1

        if dual is None:
            # we inherit the dual-ness from the axis before or after
            # to make fusing and unfusing singleton axes commutative
            if axis > 0:
                # inherit from left
                dual = self._indices[axis - 1].dual
            elif axis < self.ndim:
                # inherit from right
                dual = self._indices[axis].dual
            else:
                # no axes to inherit from
                dual = False

        if c is None:
            # insert a zero charge
            c = 0

        # insert the self charge as a column
        keys_new_col = ar.do(
            "tile", c, (self.num_blocks, 1), like=self.backend
        )
        new_sectors = ar.do(
            "concatenate",
            (self._sectors[:, :axis], keys_new_col, self._sectors[:, axis:]),
            axis=1,
            like=self.backend,
        )

        # expand the actual blocks (accounting for extra flat ax at start)
        new_blocks = ar.do(
            "expand_dims",
            self._blocks,
            axis=axis + 1,
            like=self.backend,
        )

        # expand the index information
        new_indices = (
            *self._indices[:axis],
            FlatIndex(
                num_charges=1,
                charge_size=1,
                dual=dual,
            ),
            *self._indices[axis:],
        )

        return self._modify_or_copy(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
            inplace=inplace,
        )

    def _fuse_core_abelian(
        self,
        *axes_groups,
        inplace=False,
    ) -> "FlatArrayCommon":
        """The core implementation of the fuse operation, which fuses
        multiple axes into a single group, and returns a new array with
        the new sectors and blocks. The new axes are inserted at the minimum
        axis of any of the groups.
        """
        (
            num_groups,
            group_singlets,
            new_ndim,
            _,  # perm,
            pos,
            axes_before,
            axes_after,
            _,  # ax2group,
            group_duals,
            _,  # new_axes,
        ) = calc_fuse_group_info(axes_groups, self.duals)

        # the actual number of charges in each axis (typically == order)
        axes_ncharges = tuple(ix.num_charges for ix in self._indices)

        # create the new sectors, starting with the unfused axes before
        new_sectors = [self._sectors[:, ax] for ax in axes_before]
        for axs, dg in zip(axes_groups, group_duals):
            # charges with opposite sign to overall group need to be flipped
            eff_duals = [self.duals[ax] != dg for ax in axs]
            new_sectors.append(
                zn_combine(self.order, self._sectors[:, axs], eff_duals)
            )
        # then we add the unfused axes after
        new_sectors.extend(self._sectors[:, ax] for ax in axes_after)
        # combine into single array
        new_sectors = ar.do("stack", tuple(new_sectors), axis=1)

        # then we find the correct order to sort the new keys
        sortingcols = (
            # first we sort by fused charge
            *(new_sectors[:, pos + g] for g in range(num_groups)),
            # then we sort by the unfused axes
            *(self._sectors[:, ax] for ax in axes_before),
            *(self._sectors[:, ax] for ax in axes_after),
            # and finally by the fused charges within each group
            *(self._sectors[:, ax] for group in axes_groups for ax in group),
        )
        kord = lexsort_sectors(sortingcols)
        new_blocks = self._blocks[kord]
        new_sectors = new_sectors[kord]
        # XXX: only optionally store the fusing information
        old_sectors = self._sectors[kord]

        # get the einops rearrangement pattern for the new blocks
        pattern, unmerged_batch_sizes = _calc_fuse_rearrange_pattern(
            num_groups,
            axes_groups,
            axes_before,
            axes_after,
            axes_ncharges,
            self.ndim,
        )
        # perform the rearrangement!
        new_blocks = einops_rearrange(
            new_blocks, pattern, **unmerged_batch_sizes
        )

        # now we calculate the new sectors and subkeys, either by slicing
        # the existing sectors, or by creating them from scratch
        new_sectors, subkeys = _calc_fused_sectors_subkeys_slice(
            num_groups,
            axes_groups,
            new_sectors,
            old_sectors,
            pos,
            self.ndim,
            new_ndim,
            self.order,
            group_singlets,
            self.backend,
        )
        # new_sectors, subkeys = _calc_fused_sectors_subkeys_create(
        #     unmerged_batch_sizes,
        #     num_groups,
        #     axes_groups,
        #     new_sectors,
        #     old_sectors,
        #     group_duals,
        #     self.ndim,
        #     self.order,
        #     self.duals,
        #     like=self._sectors,
        # )

        # finally we construct info, including for unfusing
        new_indices = []

        for ax in axes_before:
            new_indices.append(self.indices[ax])

        old_indices = self.indices

        for g, gaxes in enumerate(axes_groups):
            if g in group_singlets:
                # no new fuse -> just propagate current index
                new_indices.append(old_indices[gaxes[0]])
            else:
                # create a new subinfo for this group
                subindices = [old_indices[ax] for ax in gaxes]
                subinfo = FlatSubIndexInfo(subindices, subkeys[g])

                if num_groups > 1 or axes_before or axes_after:
                    num_charges = self.order
                else:
                    # output is 1D
                    num_charges = 1

                charge_size = (
                    # this is contribution from fusing the actual blocks
                    math.prod(ix.charge_size for ix in subindices)
                    # & this is contribution from grouping sectors
                    # e.g. 01 and 10 both fuse to 1
                    * self.order ** (len(gaxes) - 1)
                )

                ix = FlatIndex(
                    num_charges=num_charges,
                    charge_size=charge_size,
                    dual=group_duals[g],
                    subinfo=subinfo,
                )

                new_indices.append(ix)

        for ax in axes_after:
            new_indices.append(self.indices[ax])

        return self._modify_or_copy(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
            inplace=inplace,
        )

    def _unfuse_abelian(self, axis, inplace=False):
        new = self if inplace else self.copy()

        fi = new.indices[axis].subinfo
        if fi is None:
            raise ValueError(f"Axis {axis} is not fused in this array.")

        axs_rem = tuple(range(axis)) + tuple(range(axis + 1, self.ndim))
        new.sort_stack((axis, *axs_rem), inplace=True)

        # keys coming from remaining axes
        ka = einops_repeat(
            new.sectors, "(Bf B) s -> (Bf B x) s", Bf=fi.ncharge, x=fi.nsectors
        )[:, axs_rem]

        # keys coming from unfused axis
        kb = fi.subkeys
        kb = einops_repeat(
            kb, "B Bu s -> (B x Bu) s", x=new.num_blocks // fi.ncharge
        )

        # concatenate into the full new keys!
        new_sectors = ar.do(
            "concatenate", (ka[:, :axis], kb, ka[:, axis:]), axis=-1
        )

        # now we need to unfuse the actual blocks an example pattern:
        #     B p0 ( Bu u0 u1 u2 ) p2 p3 -> (B Bu) p0 u0 u1 u2 p2 p3
        # i.e. we unfuse the current axis, and shift its internal sector index
        # (`Bu`) into the total sector index
        pattern = ["B "]
        rhs = ["(B Bu) "]
        sizes = {}
        for i in range(axis):
            pattern.append(f"p{i} ")
            rhs.append(f"p{i} ")
        pattern.append("( Bu ")
        for g, sz in enumerate(ix.charge_size for ix in fi.indices):
            pattern.append(f"u{g} ")
            rhs.append(f"u{g} ")
            sizes[f"u{g}"] = sz
        pattern.append(") ")
        for i in range(axis + 1, new.ndim):
            pattern.append(f"p{i} ")
            rhs.append(f"p{i} ")
        pattern.append("-> ")
        pattern.extend(rhs)
        pattern = "".join(pattern)

        # perform the unfuse!
        new_blocks = einops_rearrange(new.blocks, pattern, **sizes)

        # unpack sub indices
        new_indices = (
            *new.indices[:axis],
            *fi.indices,
            *new.indices[axis + 1 :],
        )

        return new.modify(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
        )

    def select_charge(self, axis, charge, subselect=None, inplace=False):
        """Drop all but the specified charge along the specified axis. Note the
        axis is not removed, it is simply restricted to a single charge.

        Parameters
        ----------
        axis : int
            The axis along which to select the charge.
        charge : int
            The charge to select along the specified axis.
        subselect : slice or array_like, optional
            If provided, a range of indices within the selected charge block
            to keep. If not provided, the entire block is kept.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        FlatArrayCommon
        """
        if axis < 0:
            axis += self.ndim

        new = self.sort_stack(axis, inplace=inplace)

        # first we reshape the sectors and blocks so that a
        # new axis indexes the charge along the target axis
        dc = new.indices[axis].num_charges

        def _take_charge_slice(x):
            # split batch index into axis charges and slice
            xr = ar.do("reshape", x, (dc, -1, *x.shape[1:]), like=new.backend)
            return ar.do("take", xr, charge, axis=0, like=new.backend)

        new._map_blocks(
            fn_sector=_take_charge_slice,
            fn_block=_take_charge_slice,
        )

        if subselect is not None:

            def fn_block(x):
                return ar.do(
                    "take", x, subselect, axis=axis + 1, like=new.backend
                )

            # take a subselection along the selected axis
            new._map_blocks(fn_block=fn_block)

        if new.ndim == 2:
            # axes are locked to each other -> select both
            if axis == 0:
                other_charge = new.sectors[0, 1]
                new_indices = (
                    new.indices[0].select_charge(charge, subselect),
                    new.indices[1].select_charge(other_charge),
                )
            else:  # axis == 1
                other_charge = new.sectors[0, 0]
                new_indices = (
                    new.indices[0].select_charge(other_charge),
                    new.indices[1].select_charge(charge, subselect),
                )
        else:
            new_indices = (
                *new.indices[:axis],
                new.indices[axis].select_charge(charge, subselect),
                *new.indices[axis + 1 :],
            )

        return new.modify(indices=new_indices)

    def _squeeze_abelian(self, axis=None, inplace=False):
        """Squeeze the flat array, removing axes of size 1.

        Parameters
        ----------
        axis : int or sequence of int, optional
            The axes to squeeze. If not given, all axes of size 1 will be
            removed.
        inplace : bool, optional
            Whether to perform the operation inplace.

        Returns
        -------
        FlatArrayCommon
        """
        if isinstance(axis, numbers.Integral):
            axis = (axis,)

        axs_keep = []
        new_indices = []
        block_selector = [slice(None)]

        for ax, ix in enumerate(self._indices):
            if axis is None:
                remove = ix.size_total == 1
            else:
                remove = ax in axis
                if remove and ix.size_total > 1:
                    raise ValueError("Cannot squeeze d > 1 index")

            if remove:
                block_selector.append(0)
            else:
                axs_keep.append(ax)
                new_indices.append(ix)
                block_selector.append(slice(None))

        new_sectors = self.sectors[:, tuple(axs_keep)]
        new_blocks = self.blocks[tuple(block_selector)]

        return self._modify_or_copy(
            sectors=new_sectors,
            indices=tuple(new_indices),
            blocks=new_blocks,
            inplace=inplace,
        )

    def isel(self, axis, idx, inplace=False):
        """Select a single (linear) index along the specified axis. The linear
        index is first converted to the corresponding charge and offset within
        that charge sector.

        Parameters
        ----------
        axis : int
            The axis to select along.
        idx : int
            The linear index to select.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.
        """
        if axis < 0:
            axis += self.ndim
        charge, offset = self.indices[axis].linear_to_charge_and_offset(idx)
        new = self.select_charge(
            axis, charge, subselect=(offset,), inplace=inplace
        )
        # fermionic signs handled in squeeze
        return new.squeeze(axis, inplace=True)

    def align_axes(
        self: "FlatArrayCommon",
        other: "FlatArrayCommon",
        axes: tuple[tuple[int, ...], tuple[int, ...]],
        inplace=False,
    ) -> tuple["FlatArrayCommon", "FlatArrayCommon"]:
        """Align the axes of two arrays for contraction."""
        a, b = (self, other) if inplace else (self.copy(), other.copy())

        ndim_a = a.ndim
        ndim_b = b.ndim
        ncon = len(axes[0])

        if ndim_a > ncon and ndim_b > ncon:
            # ~ matmat: just need to sort sectors along common axes
            a.sort_stack(axes[0], inplace=True)
            b.sort_stack(axes[1], inplace=True)

        elif ndim_a > ncon and ndim_b == ncon:
            # ~matvec: b has locked axis and only one charge
            (axis,) = axes[0]
            a.select_charge(axis, b._sectors[0, 0], inplace=True)

        elif ndim_a == ncon and ndim_b > ncon:
            # ~vecmat: a has locked axis and only one charge
            (axis,) = axes[1]
            b.select_charge(axis, a._sectors[0, 0], inplace=True)

        else:
            # ~vecvec: must have equal charges
            matching = (a.charge + b.charge) % a.order == 0

            if a.ndim > 1:
                a.sort_stack(axes[0], inplace=True)
            if b.ndim > 1:
                b.sort_stack(axes[1], inplace=True)

            # branchless set to zero
            a.modify(blocks=a.blocks * matching)
            b.modify(blocks=b.blocks * matching)

        if DEBUG:
            a.check()
            b.check()

        return a, b

    def _tensordot_inner_abelian(
        self, other, axes_a, axes_b, preserve_array=False
    ):
        """Perform the tensor inner product of two flat abelian arrays along
        the specified axes.

        Parameters
        ----------
        other : FlatArrayCommon
            The other array to contract with.
        axes_a : tuple[int, ...]
            The axes of this array to contract along.
        axes_b : tuple[int, ...]
            The axes of the other array to contract along.
        preserve_array : bool, optional
            Whether to always return an array, even if the result is a scalar.

        Returns
        -------
        array_like | FlatArrayCommon
            The result of the contraction, either as a scalar if
            `preserve_array=False` or else a new flat abelian array.
        """
        a, b = self.align_axes(other, axes=(axes_a, axes_b))

        linput = [-1, *repeat(None, a.ndim)]
        rinput = [-1, *repeat(None, b.ndim)]
        for c, (axa, axb) in enumerate(zip(axes_a, axes_b)):
            linput[axa + 1] = c
            rinput[axb + 1] = c

        c = ctg.array_contract(
            arrays=(a.blocks, b.blocks),
            inputs=(linput, rinput),
            output=(),
        )

        if preserve_array:
            c = a.__class__.from_scalar(c, symmetry=self._symmetry)

        return c

    def _tensordot_outer_abelian(
        self, other: "FlatArrayCommon"
    ) -> "FlatArrayCommon":
        """Perform the tensor outer product of two flat abelian arrays.

        Parameters
        ----------
        other : FlatArrayCommon
            The other array to contract with.

        Returns
        -------
        FlatArrayCommon
        """
        shape_a = self._get_shape_blocks_full()
        num_blocks_a = shape_a[0]
        shape_b = other._get_shape_blocks_full()
        num_blocks_b = shape_b[0]

        # do outer via broadcasted multiplication
        new_shape_a = (num_blocks_a, 1, *shape_a[1:], *repeat(1, other.ndim))
        new_shape_b = (1, num_blocks_b, *repeat(1, self.ndim), *shape_b[1:])
        new_blocks = ar.do("reshape", self.blocks, new_shape_a) * ar.do(
            "reshape", other.blocks, new_shape_b
        )
        # remerge batch index
        new_blocks = ar.do(
            "reshape",
            new_blocks,
            (num_blocks_a * num_blocks_b, *shape_a[1:], *shape_b[1:]),
        )

        # get new keys from 'broadcasted' concatenation
        ka = einops_repeat(self.sectors, "b r -> (b x) r", x=num_blocks_b)
        kb = einops_repeat(other.sectors, "b r -> (x b) r", x=num_blocks_a)
        new_sectors = ar.do("concatenate", (ka, kb), axis=1)

        new_indices = self.indices + other.indices

        return self.new_with(
            sectors=new_sectors,
            indices=new_indices,
            blocks=new_blocks,
        )

    def _trace_abelian(self):
        """Compute the trace of the flat array, assuming it is a square
        matrix.
        """
        if self.ndim != 2:
            raise ValueError("Trace requires a 2D array.")

        _trace = ar.get_lib_fn(self.backend, "trace")
        _sum = ar.get_lib_fn(self.backend, "sum")
        new_blocks = _trace(self.blocks, axis1=1, axis2=2)
        sector_equals = self._sectors[:, 0] == self._sectors[:, 1]
        return _sum(sector_equals * new_blocks)

    def _einsum_abelian(self, eq, preserve_array=False):
        raise NotImplementedError

    def _tensordot_abelian(
        self,
        other,
        axes=2,
        mode="auto",
        preserve_array=False,
    ):
        return tensordot_abelian_flat(
            self,
            other,
            axes=axes,
            mode=mode,
            preserve_array=preserve_array,
        )

    def _matmul_abelian(
        self: "FlatArrayCommon",
        other: "FlatArrayCommon",
        preserve_array=False,
    ):
        a, b = self.align_axes(other, axes=((-1,), (0,)))

        # new sectors given by concatenation of the sectors
        new_sectors = ar.do(
            "concatenate",
            (a.sectors[:, :-1], b.sectors[:, 1:]),
            axis=1,
            like=self.backend,
        )

        inputs = []
        output = ["B"]
        if a.ndim == 1:
            inputs.append(("B", "x"))
        else:
            inputs.append(("B", "l", "x"))
            output.append("l")
        if b.ndim == 1:
            inputs.append(("B", "x"))
        else:
            inputs.append(("B", "x", "r"))
            output.append("r")

        # new full block given by batch matrix multiplication
        new_blocks = ctg.array_contract(
            arrays=(a.blocks, b.blocks),
            inputs=inputs,
            output=output,
        )

        new_indices = (*a.indices[:-1], *b.indices[1:])

        if new_indices or preserve_array:
            # array output, wrap in a new class
            return a.new_with(
                blocks=new_blocks,
                sectors=new_sectors,
                indices=new_indices,
            )

        # scalar output
        return new_blocks[0]

    def multiply_diagonal(
        self,
        v: FlatVector,
        axis,
        power=1,
        inplace=False,
    ):
        """Multiply this flat array by a vector as if contracting a diagonal
        matrix along the given axis.

        Parameters
        ----------
        v : FlatVector
            The vector to contract with.
        axis : int
            The axis along which to contract.
        inplace : bool, optional
            Whether to perform the operation inplace.

        Returns
        -------
        FlatArrayCommon
        """
        if axis < 0:
            axis += self.ndim

        # first we sort the vector by sector
        k = ar.do("argsort", v.sectors)
        vblocks_aligned = v.blocks[k]

        # expand with new broadcasting dimensions
        v_expander = (
            # sector axis we are multiplying along (aligned with blocks)
            slice(None),
            # the other sectors axis we broadcast over +
            # the data axes before axis we broadcast over
            *repeat(None, axis + 1),
            # the actual axis we are multiplying along
            slice(None),
            # the data axes after axis we broadcast over
            *repeat(None, self.ndim - axis - 1),
        )
        vblocks_aligned = vblocks_aligned[v_expander]

        # align block sectors to the vector sectors
        new = self.sort_stack(axis, inplace=inplace)
        shape_block = new.shape_block

        b_reshaper = (
            # sector axis we are multiplying along (aligned with blocks)
            new.indices[axis].num_charges,
            # other sectors axis we broadcast over
            -1,
            *shape_block,
        )
        new_blocks = ar.do("reshape", new.blocks, b_reshaper, like=new.backend)

        if power == 1:
            new_blocks = new_blocks * vblocks_aligned
        elif power == -1:
            new_blocks = new_blocks / vblocks_aligned
        else:
            # XXX: support other powers?
            raise ValueError("Invalid power value")

        # fuse axis and other sectors back into a single dimension
        new_blocks = ar.do(
            "reshape", new_blocks, (-1, *shape_block), like=new.backend
        )

        new.modify(blocks=new_blocks)
        return new

    def ldmul(self, v, inplace=False):
        return self.multiply_diagonal(v, axis=-2, inplace=inplace)

    def rdmul(self, v, inplace=False):
        return self.multiply_diagonal(v, axis=-1, inplace=inplace)

    def lddiv(self, v, inplace=False):
        return self.multiply_diagonal(v, axis=-2, power=-1, inplace=inplace)

    def rddiv(self, v, inplace=False):
        return self.multiply_diagonal(v, axis=-1, power=-1, inplace=inplace)

    def _allclose_abelian(
        self, other: "FlatArrayCommon", **allclose_opts
    ) -> bool:
        """Check if two flat abelian arrays are equal to within some tolerance,
        including their sectors and signature.
        """
        # blocks might not be stored in the same order
        a = self.sort_stack()
        b = other.sort_stack()

        if a.duals != b.duals:
            return False

        if not ar.do("allclose", a.sectors, b.sectors, **allclose_opts):
            return False

        return ar.do("allclose", a.blocks, b.blocks, **allclose_opts)

    def _test_allclose_abelian(
        self, other: "FlatArrayCommon", **allclose_opts
    ):
        """Assert that this ``FlatArayCommon`` is close to another,
        that is, has all the same sectors, and the corresponding arrays are
        close. Unlike `allclose`, this raises an AssertionError with details
        if not.

        Parameters
        ----------
        other : FlatArayCommon
            The other array to compare to.
        allclose_opts
            Keyword arguments to pass to `allclose`.

        Raises
        ------
        AssertionError
            If the arrays are not close.
        """
        # blocks might not be stored in the same order
        a = self.sort_stack()
        b = other.sort_stack()

        if a.duals != b.duals:
            raise AssertionError(
                f"Signature mismatch: {self.duals} != {other.duals}"
            )

        if not ar.do("allclose", a.sectors, b.sectors, **allclose_opts):
            raise AssertionError(
                f"Sectors not equal:\n{a.sectors}\n{b.sectors}"
            )

        if not ar.do("allclose", a.blocks, b.blocks, **allclose_opts):
            raise AssertionError(f"Blocks not equal:\n{a.blocks}\n{b.blocks}")

        return True

    # --------------------------- linalg methods ---------------------------- #

    def _qr_abelian(
        self,
        stabilized=False,
    ) -> tuple["FlatArrayCommon", "FlatArrayCommon"]:
        if self.ndim != 2:
            raise ValueError(
                "QR is only defined for 2D AbelianArrayFlat objects."
            )

        qb, rb = ar.do("linalg.qr", self._blocks, like=self.backend)

        if stabilized:
            # make each r-factor have positive diagonal
            rbd = ar.do("diagonal", rb, axis1=1, axis2=2, like=self.backend)
            r0 = rbd == 0
            s = (rbd + r0) / (ar.do("abs", rbd, like=self.backend) + r0)

            qb = qb * s[:, None, :]
            rb = rb * s[:, :, None]

        ixl, ixr = self.indices

        # drop fusing info from bond
        bond_ind = FlatIndex(
            num_charges=ixr.num_charges,
            charge_size=min(ixl.charge_size, ixr.charge_size),
            dual=ixr.dual,
        )

        q = self.copy_with(
            blocks=qb,
            indices=(ixl, bond_ind),
        )

        # R is always charge 0 and thus block diagonal
        # NOTE: we don't `copy_with` as we need to drop phases/oddpos ...
        r = self.new_with(
            sectors=self.sectors[:, (1, 1)],
            blocks=rb,
            indices=(bond_ind.conj(), ixr),
        )

        return q, r

    def _svd_abelian(
        self,
    ) -> tuple["FlatArrayCommon", FlatVector, "FlatArrayCommon"]:
        if self.ndim != 2:
            raise ValueError(
                "SVD is only defined for 2D FlatArrayCommon objects."
            )

        bu, bs, bvh = ar.do(
            "linalg.svd",
            self._blocks,
            like=self.backend,
        )

        ixl, ixr = self.indices

        # drop fusing info from bond
        bond_ind = FlatIndex(
            num_charges=ixr.num_charges,
            charge_size=min(ixl.charge_size, ixr.charge_size),
            dual=ixr.dual,
        )

        u = self.copy_with(
            blocks=bu,
            indices=(ixl, bond_ind),
        )

        s = FlatVector(
            sectors=self.sectors[:, -1],
            blocks=bs,
        )

        # VH is always charge 0 and thus block diagonal
        # NOTE: we can't `copy_with` as we need to drop phases/oddpos ...
        vh = self.new_with(
            sectors=self.sectors[:, (1, 1)],
            blocks=bvh,
            indices=(bond_ind.conj(), ixr),
        )

        if DEBUG:
            # u and vh checked in copy_with
            s.check()

        return u, s, vh

    def svd_truncated(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        **kwargs,
    ) -> tuple["FlatArrayCommon", FlatVector, "FlatArrayCommon"]:
        """Truncated singular value decomposition of this flat abelian
        symmetric array.

        Parameters
        ----------
        cutoff : float, optional
            Singular value cutoff threshold.
        cutoff_mode : int or str, optional
            How to perform the truncation:

            - 1 or 'abs': trim values below ``cutoff``
            - 2 or 'rel': trim values below ``s[0] * cutoff``
            - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

        max_bond : int
            An explicit maximum bond dimension, use -1 for none.
        absorb : {-1, 0, 1, None}
            How to absorb the singular values.

            - -1 or 'left': absorb into the left factor (U).
            - 0 or 'both': absorb the square root into both factors.
            - 1 or 'right': absorb into the right factor (VH).
            - None: do not absorb, return singular values as a BlockVector.

        renorm : {0, 1}
            Whether to renormalize the singular values (depends on
            `cutoff_mode`).
        """
        if kwargs:
            import warnings

            warnings.warn(
                f"Got unexpected kwargs {kwargs} in svd_truncated "
                f"for {self.__class__}. Ignoring them.",
                UserWarning,
            )

        # raw abelian or fermionic svd
        u, s, vh = self.svd()
        return truncate_svd_result_flat(
            u, s, vh, cutoff, cutoff_mode, max_bond, absorb, renorm
        )

    def _eigh_abelian(self) -> tuple["FlatArrayCommon", FlatVector]:
        if self.ndim != 2:
            raise NotImplementedError(
                "Eigendecomposition is only defined "
                "for 2D FlatArrayCommon objects."
            )

        eval_blocks, evec_blocks = ar.do(
            "linalg.eigh",
            self._blocks,
            like=self.backend,
        )

        eigenvectors = self.copy_with(blocks=evec_blocks)
        eigenvalues = FlatVector(
            sectors=self.sectors[:, -1], blocks=eval_blocks
        )

        if DEBUG:
            eigenvectors.check()
            eigenvalues.check()

        return eigenvalues, eigenvectors

    def _eigh_truncated_abelian(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        positive=0,
        **kwargs,
    ):
        if kwargs:
            import warnings

            warnings.warn(
                f"Got unexpected kwargs {kwargs} in eigh_truncated "
                f"for {self.__class__}. Ignoring them.",
                UserWarning,
            )

        s, U = self._eigh_abelian()

        # make sure to sort by descending absolute value
        if not positive:
            idx = ar.do(
                "argsort", -ar.do("abs", s._blocks, like=self.backend), axis=1
            )
            s.modify(
                blocks=ar.do("take_along_axis", s._blocks, idx, axis=1),
            )
            U.modify(
                blocks=ar.do(
                    "take_along_axis", U._blocks, idx[:, None, :], axis=2
                )
            )
        else:
            # assume all positive, just need to flip
            s.modify(blocks=s._blocks[:, ::-1])
            U.modify(blocks=U._blocks[:, :, ::-1])

        if DEBUG:
            s.check()
            U.check()

        # then we can truncate as if svd
        return truncate_svd_result_flat(
            U, s, U.H, cutoff, cutoff_mode, max_bond, absorb, renorm
        )


_ABSORB_MAP = {
    -1: -1,
    "left": -1,
    0: 0,
    "both": 0,
    1: 1,
    "right": 1,
    None: None,
}


def truncate_svd_result_flat(
    U: FlatArrayCommon,
    s: FlatVector,
    VH: FlatArrayCommon,
    cutoff: float,
    cutoff_mode: int,
    max_bond: int,
    absorb: int,
    renorm: int,
):
    absorb = _ABSORB_MAP[absorb]

    if cutoff > 0.0:
        raise NotImplementedError(
            "Cutoff is not implemented for flat SVD yet."
        )

    if renorm:
        raise NotImplementedError(
            "Renormalization is not implemented for flat SVD yet."
        )

    if max_bond > 0:
        # we must evenly distribute the bond dimension across charges
        bond = U.indices[1]
        # can't make bond larger
        charge_size = min(
            bond.charge_size,
            max_bond // bond.num_charges,
        )

        # we make sure to drop fusing subinfo from truncated bond
        U.modify(
            blocks=U._blocks[:, :, :charge_size],
            indices=(
                U.indices[0],
                U.indices[1].copy_with(
                    charge_size=charge_size,
                    subinfo=None,
                ),
            ),
        )
        s.modify(blocks=s._blocks[:, :charge_size])
        VH.modify(
            blocks=VH._blocks[:, :charge_size, :],
            indices=(
                VH.indices[0].copy_with(
                    charge_size=charge_size,
                    subinfo=None,
                ),
                VH.indices[1],
            ),
        )

    if absorb is None:
        if DEBUG:
            s.check()
    elif absorb == 0:
        # absorb sqrt(s) into both U and VH
        s_sqrt = s.sqrt()
        U.multiply_diagonal(s_sqrt, axis=1, inplace=True)
        VH.multiply_diagonal(s_sqrt, axis=0, inplace=True)
        s = None
    elif absorb == -1:
        # absorb s left into U
        U.multiply_diagonal(s, axis=1, inplace=True)
        s = None
    elif absorb == 1:
        # absorb s right into VH
        VH.multiply_diagonal(s, axis=0, inplace=True)
        s = None
    else:
        raise ValueError(
            f"absorb must be 0, -1, 1, or None. Got {absorb} instead."
        )

    if DEBUG:
        U.check()
        VH.check()

    return U, s, VH


# ------------------------ tensordot implementations ------------------------ #


def tensordot_flat_fused(
    a: FlatArrayCommon,
    b: FlatArrayCommon,
    left_axes: tuple[int, ...],
    axes_a: tuple[int, ...],
    axes_b: tuple[int, ...],
    right_axes: tuple[int, ...],
    preserve_array=False,
):
    # NOTE: we use abelian fusion as, while fermionic tensordot uses this
    # routine, it handles all phases ahead of time then expects abelian

    if left_axes:
        af = a._fuse_core_abelian(left_axes, axes_a)
    elif len(axes_a) > 1:
        af = a._fuse_core_abelian(axes_a)
    else:
        af = a

    if right_axes:
        bf = b._fuse_core_abelian(axes_b, right_axes)
    elif len(axes_b) > 1:
        bf = b._fuse_core_abelian(axes_b)
    else:
        bf = b

    cf = af._matmul_abelian(bf, preserve_array=preserve_array)

    if isinstance(cf, FlatArrayCommon):
        # if we got a new flat array, unfuse all axes
        for ax in reversed(range(cf.ndim)):
            if cf.is_fused(ax):
                cf = cf._unfuse_abelian(ax, inplace=True)

    return cf


def tensordot_flat_direct(
    a: FlatArrayCommon,
    b: FlatArrayCommon,
    left_axes: tuple[int, ...],
    axes_a: tuple[int, ...],
    axes_b: tuple[int, ...],
    right_axes: tuple[int, ...],
    preserve_array=False,
):
    """Contract two flat abelian arrays without fusing."""
    if not left_axes or not right_axes:
        raise NotImplementedError(
            "Currently both `a` and `b` must have kept axes for direct "
            "tensordot. Consider using mode='fused' instead."
        )

    _reshape = ar.get_lib_fn(a.backend, "reshape")

    dc = a.order ** (len(axes_a) - 1)

    # sort and reshape left blocks
    d0 = a.duals[axes_a[0]]
    lcon_sectors = zn_combine(
        a.order,
        a.sectors[:, axes_a],
        duals=[a.duals[ax] != d0 for ax in axes_a],
    )
    lkord = lexsort_sectors(
        (
            lcon_sectors,
            *(a.sectors[:, ax] for ax in left_axes),
            *(a.sectors[:, ax] for ax in axes_a),
        )
    )
    larray = _reshape(a.blocks[lkord], (a.order, -1, dc, *a.shape_block))

    # sort and reshape right blocks
    d0 = b.duals[axes_b[0]]
    rcon_sectors = zn_combine(
        b.order,
        b.sectors[:, axes_b],
        duals=[b.duals[ax] != d0 for ax in axes_b],
    )
    rkord = lexsort_sectors(
        (
            rcon_sectors,
            *(b.sectors[:, ax] for ax in right_axes),
            *(b.sectors[:, ax] for ax in axes_b),
        )
    )
    rarray = _reshape(b.blocks[rkord], (b.order, -1, dc, *b.shape_block))

    linput = ["B0", "Bl", "Bc"]
    rinput = ["B0", "Br", "Bc"]
    output = ["B0", "Bl", "Br"]

    aconmap = {}
    bconmap = {}
    for c, (axa, axb) in enumerate(zip(axes_a, axes_b)):
        aconmap[int(axa)] = c
        bconmap[int(axb)] = c

    for i in range(a.ndim):
        c = aconmap.get(i, None)
        if c is None:
            linput.append(f"a{i}")
            output.append(f"a{i}")
        else:
            linput.append(f"C{c}")

    for i in range(b.ndim):
        c = bconmap.get(i, None)
        if c is None:
            rinput.append(f"b{i}")
            output.append(f"b{i}")
        else:
            rinput.append(f"C{c}")

    new_blocks = ctg.array_contract(
        (larray, rarray), inputs=(linput, rinput), output=output
    )

    # flatten all batch indices
    db, dl, dr, *shape_new_block = new_blocks.shape
    new_blocks = _reshape(new_blocks, (db * dl * dr, *shape_new_block))

    # XXX: do we need the B axis at all here?

    # now we handle sectors
    lsectors = _reshape(a.sectors[lkord], (a.order, -1, dc, a.ndim))[
        :, :, 0, left_axes
    ]
    rsectors = _reshape(b.sectors[rkord], (b.order, -1, dc, b.ndim))[
        :, :, 0, right_axes
    ]

    new_sectors = ar.do(
        "concatenate",
        (
            einops_repeat(lsectors, "B Bl c -> (B Bl repeat) c", repeat=dr),
            einops_repeat(rsectors, "B Br c -> (B repeat Br) c", repeat=dl),
        ),
        axis=1,
    )

    new_indices = (
        *(a.indices[ax] for ax in left_axes),
        *(b.indices[ax] for ax in right_axes),
    )

    if new_indices or preserve_array:
        # array output, wrap in a new class
        return a.new_with(
            sectors=new_sectors,
            blocks=new_blocks,
            indices=new_indices,
        )

    # scalar output
    return new_blocks[0]


def tensordot_abelian_flat(
    a: FlatArrayCommon,
    b: FlatArrayCommon,
    axes=2,
    mode="auto",
    preserve_array=False,
):
    """Contract two flat abelian arrays along the specified axes."""
    left_axes, axes_a, axes_b, right_axes = parse_tensordot_axes(
        axes, a.ndim, b.ndim
    )

    if not axes_a:
        # outer product
        return a._tensordot_outer_abelian(b)

    if not (left_axes or right_axes):
        # inner product
        return a._tensordot_inner_abelian(b, axes_a, axes_b, preserve_array)

    if mode == "auto":
        if left_axes and right_axes:
            mode = "direct"
        else:
            mode = "fused"

    if mode == "direct":
        return tensordot_flat_direct(
            a,
            b,
            left_axes=left_axes,
            axes_a=axes_a,
            axes_b=axes_b,
            right_axes=right_axes,
            preserve_array=preserve_array,
        )
    elif mode == "fused":
        return tensordot_flat_fused(
            a,
            b,
            left_axes=left_axes,
            axes_a=axes_a,
            axes_b=axes_b,
            right_axes=right_axes,
            preserve_array=preserve_array,
        )
    else:
        raise ValueError(f"Unknown flat tensordot mode: {mode}")


def print_charge_fusions(keys, duals, axes_groups):
    from colorama import Fore

    RESET = Fore.RESET
    PINK = Fore.MAGENTA
    RED = Fore.RED
    BLUE = Fore.BLUE
    GREEN = Fore.GREEN

    color_options = [BLUE, RED, GREEN, PINK]

    (
        num_groups,
        _,  # group_singlets,
        new_ndim,
        _,  # perm,
        position,
        axes_before,
        axes_after,
        ax2group,
        _,  # group_duals,
        new_axes,
    ) = calc_fuse_group_info(axes_groups, duals)

    colors = dict(zip(range(num_groups), color_options))
    colors[None] = RESET
    old_key = None

    keys = ar.do("reshape", keys, (-1, ar.do("shape", keys)[-1]))

    for key in keys:
        new_key = [0] * new_ndim
        for oldax, newax in new_axes.items():
            new_key[newax] ^= int(key[oldax])

        if old_key is not None and old_key != new_key:
            print("- " * len(duals))
        old_key = new_key

        for i, c in enumerate(key):
            g = ax2group[i]
            print(f"{colors[g]}{c}{RESET}", end=" ")
        print("->", end=" ")

        for ax in axes_before:
            print(f"{key[ax]}", end=" ")
        for g in range(num_groups):
            print(f"{colors[g]}{new_key[position + g]}{RESET}", end=" ")
        for ax in axes_after:
            print(f"{key[ax]}", end=" ")

        print()


def build_cyclic_keys_all(ndim, order=2, flat=False, like=None):
    """For cyclic group of order `order`, build all possible subkeys of
    length `ndim`, in lexicographic order.

    Parameters
    ----------
    ndim : int
        The number of sub charges to build keys for.
    order : int, optional
        The order of the cyclic group, i.e., the number of distinct charges.
        Default is 2, which corresponds to the Z2 group.
    flat : bool, optional
        Whether to flatten the keys into a 2D array. If True, the output will
        be a 2D array of shape (order ** ndim, ndim). If False, the output will
        be a multi-dimensional array of shape `(order,) * ndim + (ndim,)`.
        Default is False.
    like : array_like, optional
        If provided, the output will be created with the same backend as this
        array. If None, the output will be created with the default backend,
        which is usually `numpy`.

    Returns
    -------
    array_like
        An array of shape (order ** ndim, ndim) or
        (order, order, ..., order, ndim) depending on the `flat` parameter.
    """
    kws = {}
    if like is None:
        kws["dtype"] = "int64"
    keys = ar.do("zeros", (order,) * ndim + (ndim,), like=like, **kws)

    for i in range(1, order):
        for j in range(ndim):
            selector = (
                *(
                    slice(i, i + 1) if k == j else slice(None)
                    for k in range(ndim)
                ),
                j,
            )
            keys[selector] = i

    if flat:
        # flatten the keys to a 2D array
        keys = ar.do("reshape", keys, (-1, ndim), like=like)

    return keys


def build_cyclic_keys_conserve(
    ndim,
    order=2,
    charge=0,
    duals=None,
    flat=False,
    like=None,
):
    """For cyclic group of order `order`, build all possible subkeys of
    length `ndim` with overall charge `charge`, in lexicographic order.

    Parameters
    ----------
    ndim : int
        The number of sub charges to build keys for.
    order : int, optional
        The order of the cyclic group, i.e., the number of distinct charges.
        Default is 2, which corresponds to the Z2 group.
    flat : bool, optional
        Whether to flatten the keys into a 2D array. If True, the output will
        be a 2D array of shape (order ** (ndim - 1), ndim). If False, the
        output will be a multi-dimensional array of shape
        `(order,) * (ndim - 1) + (ndim,)`.. Default is False.
    like : array_like, optional
        If provided, the output will be created with the same backend as this
        array. If None, the output will be created with the default backend,
        which is usually `numpy`.

    Returns
    -------
    array_like
        An array of shape (order ** (ndim - 1), ndim) or
        (order,) * (ndim - 1) + (ndim,) depending on the `flat` parameter.
    """
    kws = {}
    if like is None:
        kws["dtype"] = "int64"
    keys = ar.do("zeros", (order,) * (ndim - 1) + (ndim,), like=like, **kws)

    for i in range(1, order):
        for j in range(ndim - 1):
            selector = (
                *(
                    slice(i, i + 1) if k == j else slice(None)
                    for k in range(ndim - 1)
                ),
                j,
            )
            keys[selector] = i

    if duals is not None:
        signs = ar.do("array", [-1 if d else 1 for d in duals], like=keys)
        last_sign = signs[-1]
        signs = ar.do("reshape", signs, (1,) * ndim + (-1,), like=keys)
        keys[..., -1] = (
            last_sign * (charge - ar.do("sum", signs * keys, axis=-1))
        ) % order
    else:
        keys[..., -1] = (charge - (ar.do("sum", keys, axis=-1))) % order

    if flat:
        # flatten the keys to a 2D array
        keys = ar.do("reshape", keys, (-1, ndim), like=like)

    return keys


def build_cyclic_keys_by_charge(ndim, order=2, duals=None, like=None):
    """For cyclic group of order `order`, build all possible subkeys of
    length `ndim`, grouped (via the first axis) by their overall charge,
    then lexicographically ordered within each charge.

    Parameters
    ----------
    ndim : int
        The number of sub charges to build keys for.
    order : int, optional
        The order of the cyclic group, i.e., the number of distinct charges.
        Default is 2, which corresponds to the Z2 group.
    like : array_like, optional
        If provided, the output will be created with the same backend as this
        array. If None, the output will be created with the default backend,
        which is usually `numpy`.

    Returns
    -------
    array_like
        An array of shape (order, order ** (ndim - 1), ndim).
    """
    kws = {}
    if like is None:
        kws["dtype"] = "int64"
    keys = ar.do("zeros", (order,) * ndim + (ndim,), like=like, **kws)

    for i in range(1, order):
        for j in range(ndim - 1):
            selector = (
                slice(None),
                *(
                    slice(i, i + 1) if k == j else slice(None)
                    for k in range(ndim - 1)
                ),
                j,
            )
            keys[selector] = i

    # flatten all but the overall charge and sector axis
    keys = ar.do("reshape", keys, (order, -1, ndim), like=like)

    # create broadcastable array of possible total charges
    c_total = ar.do("reshape", ar.do("arange", order), (order, 1), like=like)

    # compute last column of each sector, based on the total charge
    if duals is not None:
        signs = ar.do("array", [-1 if d else 1 for d in duals], like=keys)
        last_sign = signs[-1]
        signs = ar.do("reshape", signs, (1,) * (ndim + 1) + (-1,), like=keys)
        c_last = (
            last_sign
            * (c_total - ar.do("sum", signs * keys, axis=-1, like=like))
        ) % order
    else:
        c_last = (c_total - ar.do("sum", keys, axis=-1, like=like)) % order
    keys[:, :, -1] = c_last

    return keys



"""Abelian symmetric arrays with flat backend.

Flat backend with branchless design to allow static
computational graphs that can be easily compiled and vectorized etc.

TODO:

- [ ] cache patterns and reshapers/slicers
- [ ] cache properties, funcs
- [ ] roll our own repeat and rearrange to avoid einops dependency

"""

import autoray as ar

from ..abelian_common import AbelianCommon
from ..common import SymmrayCommon
from ..sparse.sparse_abelian_array import AbelianArray
from ..symmetries import get_symmetry
from ..utils import DEBUG
from .flat_array_common import FlatArrayCommon
from .flat_data_common import FlatCommon
from .flat_vector import FlatVector


class AbelianArrayFlat(
    FlatArrayCommon,
    FlatCommon,
    AbelianCommon,
    SymmrayCommon,
):
    """Base class for abelian arrays with flat storage and cyclic symmetry.

    Parameters
    ----------
    sectors : array_like
        The stack of sector keys, with shape (num_blocks, ndim). Each row
        represents a sector of a corresponding block, and each column
        represents a charge in a given axis.
    blocks : array_like
        The stack of array blocks, with shape (num_blocks, *shape_block), i.e.
        `ndim + 1` dimensions, where the first dimension is the block index,
        which should match the first dimension of `sectors`, and the rest are
        the dimensions of individual blocks.
    indices : sequence[FlatIndex]
        Indices describing the dualness and any subindex information for each
        dimension of the array. If bools are supplied, they will be converted
        to a FlatIndex with the corresponding dualness, and no subindex
        information.
    symmetry : str or Symmetry, optional
        The symmetry of the array, if not using a specific symmetry class.
    """

    __slots__ = (
        "_blocks",
        "_indices",
        "_label",
        "_sectors",
        "_symmetry",
        "backend",
    )

    fermionic = False
    static_symmetry = None

    def __init__(
        self,
        sectors,
        blocks,
        indices,
        symmetry=None,
        label=None,
    ):
        self._init_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
            symmetry=symmetry,
            label=label,
        )

        if DEBUG:
            # might not have completed fermionic setup yet
            self.check()

    def check(self):
        """Check the internal consistency of the array."""
        self._check_abelian()

    def new_with(self, sectors, blocks, indices) -> "AbelianArrayFlat":
        """Create a new flat abelian array of the same class as this one.
        Unlike `copy`, this does not copy over any existing data and drops
        for example `label`.
        """
        return self._new_with_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
        )

    def copy(self, deep=False) -> "AbelianArrayFlat":
        """Create a copy of the array."""
        return self._copy_abelian(deep=deep)

    def copy_with(
        self,
        sectors=None,
        blocks=None,
        indices=None,
    ) -> "AbelianArrayFlat":
        """A copy of this block array with some attributes replaced. Note that
        checks are not performed on the new properties, this is intended for
        internal use.
        """
        new = self._copy_with_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
        )
        if DEBUG:
            new.check()
        return new

    def modify(
        self,
        sectors=None,
        blocks=None,
        indices=None,
    ) -> "AbelianArrayFlat":
        """Modify this flat array in place with some attributes replaced. Note
        that checks are not performed on the new properties, this is intended
        for internal use.
        """
        self._modify_abelian(
            sectors=sectors,
            blocks=blocks,
            indices=indices,
        )
        if DEBUG:
            self.check()
        return self

    def set_params(self, params):
        """Set the underlying array blocks."""
        self._set_params_abelian(params)
        if DEBUG:
            self.check()

    @classmethod
    def from_blocks(
        cls, blocks, indices, symmetry=None, label=None
    ) -> "AbelianArrayFlat":
        """Create a flat array from an explicit dictionary of blocks, and
        sequence of indices or duals.

        Parameters
        ----------
        blocks : dict[tuple[int, ...], array_like]
            A dictionary mapping sector keys (tuples of charges) to blocks
            (arrays).
        indices : sequence[FlatIndex] | sequence[bool]
            A sequence of indices describing the dualness and any subindex
            information for each dimension of the array. If bools are supplied,
            they will be converted to a FlatIndex with the corresponding
            dualness, and no subindex information.
        symmetry : str or Symmetry, optional
            The symmetry of the array, if not using a specific symmetry class.
        label : hashable, optional
            An optional label for the array, potentially needed for ordering
            dummy odd fermionic modes.


        Returns
        -------
        AbelianArrayFlat
        """
        sectors = list(map(list, blocks.keys()))

        if blocks:
            blocks = ar.do("stack", tuple(blocks.values()))
        else:
            blocks = []
        return cls(
            sectors,
            blocks,
            indices,
            symmetry=symmetry,
            label=label,
        )

    @classmethod
    def from_blocksparse(
        cls,
        x: AbelianArray,
        symmetry=None,
    ) -> "AbelianArrayFlat":
        """Create a flat abelian array from a blocksparse abelian array.

        Parameters
        ----------
        x : AbelianArray
            The blocksparse abelian array to convert.
        symmetry : str or Symmetry, optional
            The symmetry to use. If not supplied, the symmetry of `x` is used.
        """
        # ensure we only have index charges present on blocks
        x = x.sync_charges()

        return cls.from_blocks(
            blocks=x.blocks,
            indices=[ix.to_flat() for ix in x.indices],
            symmetry=symmetry or x.symmetry,
            label=x.label,
        )

    def to_blocksparse(self) -> AbelianArray:
        """Create a blocksparse abelian array from this flat abelian array."""
        return self._to_blocksparse_abelian()

    def to_dense(self):
        """Convert this abelian array to a dense array, by combining all the
        blocks into a single large array, filling in zeros where necessary.

        Returns
        -------
        array_like
            A dense array with the same shape as this abelian array.
        """
        return self._to_dense_abelian()

    def _map_blocks(self, fn_sector=None, fn_block=None):
        self._map_blocks_abelian(fn_sector=fn_sector, fn_block=fn_block)

    def sort_stack(
        self,
        axes=None,
        all_axes=None,
        inplace=False,
    ) -> "AbelianArrayFlat":
        """Lexicgraphic sort the stack of sectors and blocks according to the
        values of charges in the specified axes, optionally filling in the rest
        of the axes with the remaining axes in the order they appear.

        Parameters
        ----------
        axes : int | tuple[int, ...], optional
            The axes to sort by. If a single integer is given, it will be
            interpreted as the axis to sort by. If a tuple of integers is
            given, it will be interpreted as the axes to sort by in order.
            Default is None, if all_axes is also None or True, this will sort
            all axes in their current order.
        all_axes : bool, optional
            Whether to include all non-specified axes as tie-breakers, after
            the specified axes. If ``None``, the default, this will be True
            if `axes` is not supplied explicitly, and False otherwise.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.
            Default is False, which returns a new array.
        """
        kord = self.get_sorting_indices(axes=axes, all_axes=all_axes)
        return self._modify_or_copy(
            sectors=self._sectors[kord],
            blocks=self._blocks[kord],
            inplace=inplace,
        )

    def _binary_blockwise_op(self, other, fn, missing=None, inplace=False):
        return self._binary_blockwise_op_abelian(
            other, fn, missing=missing, inplace=inplace
        )

    def transpose(
        self,
        axes=None,
        inplace=False,
    ) -> "AbelianArrayFlat":
        """Transpose this flat abelian array.

        Parameters
        ----------
        axes : tuple[int, ...] | None, optional
            A permutation of the axes to transpose the array by. If None,
            the axes will be reversed.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArrayFlat
        """
        return self._transpose_abelian(axes=axes, inplace=inplace)

    def conj(self, inplace=False) -> "AbelianArrayFlat":
        """Return the complex conjugate of this flat abelian array, including
        the indices and any subindex fusing information.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArrayFlat
        """
        return self._conj_abelian(inplace=inplace)

    def dagger(self, inplace=False) -> "AbelianArrayFlat":
        """Return the adjoint of this flat abelian array, including the
        indices and any subindex fusing information.

        Parameters
        ----------
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArrayFlat
        """
        return self._dagger_abelian(inplace=inplace)

    def _fuse_core(
        self,
        *axes_groups,
        inplace=False,
    ) -> "AbelianArrayFlat":
        """The core implementation of the fuse operation, which fuses
        multiple axes into a single group, and returns a new array with
        the new sectors and blocks. The new axes are inserted at the minimum
        axis of any of the groups.
        """
        return self._fuse_core_abelian(*axes_groups, inplace=inplace)

    def unfuse(self, axis, inplace=False) -> "AbelianArrayFlat":
        """Unfuse the ``axis`` index, which must carry subindex information,
        likely generated automatically from a fusing operation.

        Parameters
        ----------
        axis : int
            The axis to unfuse. It must have fuse information
            (`.indices[axis].subinfo`), typically from a previous fusing
            operation.
        inplace : bool, optional
            Whether to perform the operation inplace or return a new array.

        Returns
        -------
        AbelianArrayFlat
        """
        return self._unfuse_abelian(axis, inplace=inplace)

    def __matmul__(
        self: "AbelianArrayFlat",
        other: "AbelianArrayFlat",
        preserve_array=False,
    ):
        return self._matmul_abelian(other=other, preserve_array=preserve_array)

    def tensordot(
        self,
        other,
        axes=2,
        mode="auto",
        preserve_array=False,
    ):
        return self._tensordot_abelian(
            other, axes=axes, mode=mode, preserve_array=preserve_array
        )

    def trace(self):
        """Compute the trace of the flat array, assuming it is a square matrix."""
        return self._trace_abelian()

    def einsum(self, eq, preserve_array=False):
        """Einsum for flat abelian arrays, currently only single term.

        Parameters
        ----------
        eq : str
            The einsum equation, e.g. "abcb->ca". The output indices must be
            specified and only trace and permutations are allowed.
        preserve_array : bool, optional
            If tracing to a scalar, whether to return an AbelainArray object
            with no indices, or simply scalar itself (the default).

        Returns
        -------
        FlatAbelianArray or scalar
        """
        return self._einsum_abelian(eq, preserve_array=preserve_array)

    def squeeze(self, axis, inplace=False):
        """Squeeze this flat abelian array, removing axes of size 1.

        Parameters
        ----------
        axis : int or sequence of int, optional
            The axes to squeeze. If not given, all axes of size 1 will be
            removed.
        inplace : bool, optional
            Whether to perform the operation inplace.

        Returns
        -------
        AbelianArrayFlat
        """
        return self._squeeze_abelian(axis, inplace=inplace)

    def allclose(self, other: "AbelianArrayFlat", **allclose_opts) -> bool:
        """Check if two flat abelian arrays are equal to within some tolerance,
        including their sectors and signature.
        """
        return self._allclose_abelian(other, **allclose_opts)

    def test_allclose(self, other: "AbelianArrayFlat", **allclose_opts):
        """Assert that this ``AbelianArrayFlat`` is close to another,
        that is, has all the same sectors, and the corresponding arrays are
        close. Unlike `allclose`, this raises an AssertionError with details
        if not.

        Parameters
        ----------
        other : AbelianArrayFlat
            The other array to compare to.
        allclose_opts
            Keyword arguments to pass to `allclose`.

        Raises
        ------
        AssertionError
            If the arrays are not close.
        """
        return self._test_allclose_abelian(other, **allclose_opts)

    # --------------------------- linalg methods ---------------------------- #

    def qr(
        self,
        stabilized=False,
    ) -> tuple["AbelianArrayFlat", "AbelianArrayFlat"]:
        """QR decomposition of this flat abelian array.

        Parameters
        ----------
        x : AbelianArrayFlat
            The flat symmetric array to decompose.
        stabilized : bool, optional
            Whether to use a stabilized QR decomposition, that is, with
            positive diagonal elements in the R factor. Default is False.

        Returns
        -------
        q : AbelianArrayFlat
            The orthogonal matrix.
        r : AbelianArrayFlat
            The upper triangular matrix.
        """
        return self._qr_abelian(stabilized=stabilized)

    def svd(self) -> tuple["AbelianArrayFlat", FlatVector, "AbelianArrayFlat"]:
        """Singular value decomposition of this flat abelian array.

        Returns
        -------
        u : AbelianArrayFlat
            The left singular vectors.
        s : FlatVector
            The singular values.
        vh : AbelianArrayFlat
            The right singular vectors (hermitian transposed).
        """
        return self._svd_abelian()

    def eigh(self) -> tuple[FlatVector, "AbelianArrayFlat"]:
        """Hermitian eigen-decomposition of this flat abelian array.

        Returns
        -------
        eigenvalues : FlatVector
            The eigenvalues.
        eigenvectors : AbelianArrayFlat
            The abelian array of right eigenvectors.
        """
        return self._eigh_abelian()

    def eigh_truncated(
        self,
        cutoff=-1.0,
        cutoff_mode=4,
        max_bond=-1,
        absorb=0,
        renorm=0,
        positive=0,
        **kwargs,
    ) -> tuple["AbelianArrayFlat", FlatVector, "AbelianArrayFlat"]:
        """Truncated hermitian eigen-decomposition of this assumed hermitian
        flat abelian array.

        Parameters
        ----------
        cutoff : float, optional
            Absolute eigenvalue cutoff threshold.
        cutoff_mode : int or str, optional
            How to perform the truncation:

            - 1 or 'abs': trim values below ``cutoff``
            - 2 or 'rel': trim values below ``s[0] * cutoff``
            - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
            - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
            - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
            - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

        max_bond : int
            An explicit maximum bond dimension, use -1 for none.
        absorb : {-1, 0, 1, None}
            How to absorb the eigenvalues.

            - -1 or 'left': absorb into the left factor (U).
            - 0 or 'both': absorb the square root into both factors.
            - 1 or 'right': absorb into the right factor (VH).
            - None: do not absorb, return eigenvalues as a BlockVector.

        renorm : {0, 1}
            Whether to renormalize the eigenvalues (depends on `cutoff_mode`).

        Returns
        -------
        u : AbelianArrayFlat
            The abelian array of left eigenvectors.
        w : VectorCommon or None
            The vector of eigenvalues, or None if absorbed.
        uh : AbelianArrayFlat
            The abelian array of right eigenvectors.
        """
        return self._eigh_truncated_abelian(
            cutoff=cutoff,
            cutoff_mode=cutoff_mode,
            max_bond=max_bond,
            absorb=absorb,
            renorm=renorm,
            positive=positive,
            **kwargs,
        )


class Z2ArrayFlat(AbelianArrayFlat):
    static_symmetry = get_symmetry("Z2")
